@@ -15,8 +15,9 @@
  */
 #include "config.h"
 #include "libutil/util.h"
-#include "libutil/http.h"
-#include "libutil/http_private.h"
+#include "libserver/http/http_connection.h"
+#include "libserver/http/http_private.h"
+#include "libserver/cfg_file.h"
 #include "rspamdclient.h"
 #include "utlist.h"
 #include "unix-std.h"
@@ -67,15 +68,22 @@ static gchar *key = NULL;
 static gchar *user_agent = "rspamc";
 static GList *children;
 static GPatternSpec **exclude_compiled = NULL;
+static struct rspamd_http_context *http_ctx;
+
+static gint retcode = EXIT_SUCCESS;
 
 #define ADD_CLIENT_HEADER(o, n, v) do { \
     struct rspamd_http_client_header *nh; \
     nh = g_malloc (sizeof (*nh)); \
-    nh->name = (n); \
-    nh->value = (v); \
+    nh->name = g_strdup (n); \
+    nh->value = g_strdup (v); \
     g_queue_push_tail ((o), nh); \
 } while (0)
 
+#define ADD_CLIENT_FLAG(str, n) do { \
+   g_string_append ((str), n ","); \
+} while (0)
+
 static gboolean rspamc_password_callback (const gchar *option_name,
 		const gchar *value,
 		gpointer data,
@@ -161,17 +169,6 @@ static GOptionEntry entries[] =
 	{ NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL }
 };
 
-/* Copy to avoid linking with librspamdserver */
-enum rspamd_action_type {
-	METRIC_ACTION_REJECT = 0,
-	METRIC_ACTION_SOFT_REJECT,
-	METRIC_ACTION_REWRITE_SUBJECT,
-	METRIC_ACTION_ADD_HEADER,
-	METRIC_ACTION_GREYLIST,
-	METRIC_ACTION_NOACTION,
-	METRIC_ACTION_MAX
-};
-
 static void rspamc_symbols_output (FILE *out, ucl_object_t *obj);
 static void rspamc_uptime_output (FILE *out, ucl_object_t *obj);
 static void rspamc_counters_output (FILE *out, ucl_object_t *obj);
@@ -404,6 +401,7 @@ read_cmd_line (gint *argc, gchar ***argv)
 	/* Parse options */
 	if (!g_option_context_parse (context, argc, argv, &error)) {
 		fprintf (stderr, "option parsing failed: %s\n", error->message);
+		g_option_context_free (context);
 		exit (EXIT_FAILURE);
 	}
 
@@ -411,6 +409,7 @@ read_cmd_line (gint *argc, gchar ***argv)
 		raw = TRUE;
 	}
 	/* Argc and argv are shifted after this function */
+	g_option_context_free (context);
 }
 
 static gboolean
@@ -550,15 +549,13 @@ add_options (GQueue *opts)
 {
 	GString *numbuf;
 	gchar **hdr, **rcpt;
-
-	if (user_agent) {
-		ADD_CLIENT_HEADER (opts, "User-Agent", user_agent);
-	}
+	GString *flagbuf = g_string_new (NULL);
 
 	if (ip != NULL) {
 		rspamd_inet_addr_t *addr = NULL;
 
-		if (!rspamd_parse_inet_address (&addr, ip, strlen (ip))) {
+		if (!rspamd_parse_inet_address (&addr, ip, strlen (ip),
+				RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
 			/* Try to resolve */
 			struct addrinfo hints, *res, *cur;
 			gint r;
@@ -635,7 +632,7 @@ add_options (GQueue *opts)
 	}
 
 	if (pass_all) {
-		ADD_CLIENT_HEADER (opts, "Pass", "all");
+		ADD_CLIENT_FLAG (flagbuf, "pass_all");
 	}
 
 	if (classifier) {
@@ -646,6 +643,7 @@ add_options (GQueue *opts)
 		numbuf = g_string_sized_new (8);
 		rspamd_printf_gstring (numbuf, "%d", weight);
 		ADD_CLIENT_HEADER (opts, "Weight", numbuf->str);
+		g_string_free (numbuf, TRUE);
 	}
 
 	if (fuzzy_symbol != NULL) {
@@ -656,6 +654,7 @@ add_options (GQueue *opts)
 		numbuf = g_string_sized_new (8);
 		rspamd_printf_gstring (numbuf, "%d", flag);
 		ADD_CLIENT_HEADER (opts, "Flag", numbuf->str);
+		g_string_free (numbuf, TRUE);
 	}
 
 	if (extended_urls) {
@@ -663,9 +662,11 @@ add_options (GQueue *opts)
 	}
 
 	if (profile) {
-		ADD_CLIENT_HEADER (opts, "Profile", "true");
+		ADD_CLIENT_FLAG (flagbuf, "profile");
 	}
 
+	ADD_CLIENT_FLAG (flagbuf, "body_block");
+
 	if (skip_images) {
 		ADD_CLIENT_HEADER (opts, "Skip-Images", "true");
 	}
@@ -681,17 +682,30 @@ add_options (GQueue *opts)
 
 		if (kv == NULL || kv[1] == NULL) {
 			ADD_CLIENT_HEADER (opts, *hdr, "");
-
-			if (kv) {
-				g_strfreev (kv);
-			}
 		}
 		else {
 			ADD_CLIENT_HEADER (opts, kv[0], kv[1]);
 		}
 
+		if (kv) {
+			g_strfreev (kv);
+		}
+
 		hdr ++;
 	}
+
+	if (flagbuf->len > 0) {
+		goffset last = flagbuf->len - 1;
+
+		if (flagbuf->str[last] == ',') {
+			flagbuf->str[last] = '\0';
+			flagbuf->len --;
+		}
+
+		ADD_CLIENT_HEADER (opts, "Flags", flagbuf->str);
+	}
+
+	g_string_free (flagbuf, TRUE);
 }
 
 static void
@@ -885,7 +899,15 @@ rspamc_symbols_output (FILE *out, ucl_object_t *obj)
 		}
 	}
 
-	PRINT_PROTOCOL_STRING ("dkim-signature", "DKIM-Signature");
+	elt = ucl_object_lookup (obj, "dkim-signature");
+	if (elt && elt->type == UCL_STRING) {
+		rspamd_fprintf (out, "DKIM-Signature: %s\n", ucl_object_tostring (elt));
+	} else if (elt && elt->type == UCL_ARRAY) {
+		mit = NULL;
+		while ((cmesg = ucl_object_iterate (elt, &mit, true)) != NULL) {
+			rspamd_fprintf (out, "DKIM-Signature: %s\n", ucl_object_tostring (cmesg));
+		}
+	}
 
 	elt = ucl_object_lookup (obj, "profile");
 
@@ -999,9 +1021,9 @@ rspamc_counters_sort (const ucl_object_t **o1, const ucl_object_t **o2)
 					order2 = ucl_object_toint (elt2);
 				}
 			}
-
-			g_strfreev (args);
 		}
+
+		g_strfreev (args);
 	}
 
 	return (inverse ? (order2 - order1) : (order1 - order2));
@@ -1013,7 +1035,6 @@ rspamc_counters_output (FILE *out, ucl_object_t *obj)
 	const ucl_object_t *cur, *sym, *weight, *freq, *freq_dev, *nhits;
 	ucl_object_iter_t iter = NULL;
 	gchar fmt_buf[64], dash_buf[82], sym_buf[82];
-	gint l, max_len = INT_MIN, i;
 	static const gint dashes = 44;
 
 	if (obj->type != UCL_ARRAY) {
@@ -1027,16 +1048,17 @@ rspamc_counters_output (FILE *out, ucl_object_t *obj)
 	}
 
 	/* Find maximum width of symbol's name */
+	gint max_len = sizeof("Symbol") - 1;
 	while ((cur = ucl_object_iterate (obj, &iter, true)) != NULL) {
 		sym = ucl_object_lookup (cur, "symbol");
 		if (sym != NULL) {
-			l = sym->len;
-			if (l > max_len) {
-				max_len = MIN (sizeof (dash_buf) - dashes - 1, l);
+			if (sym->len > max_len) {
+				max_len = sym->len;
 			}
 		}
 	}
 
+	max_len = MIN (sizeof (dash_buf) - dashes - 1, max_len);
 	rspamd_snprintf (fmt_buf, sizeof (fmt_buf),
 		"| %%3s | %%%ds | %%7s | %%13s | %%7s |\n", max_len);
 	memset (dash_buf, '-', dashes + max_len);
@@ -1057,7 +1079,7 @@ rspamc_counters_output (FILE *out, ucl_object_t *obj)
 		"| %%3d | %%%ds | %%7.1f | %%6.3f(%%5.3f) | %%7ju |\n", max_len);
 
 	iter = NULL;
-	i = 0;
+	gint i = 0;
 	while ((cur = ucl_object_iterate (obj, &iter, true)) != NULL) {
 		printf (" %s \n", dash_buf);
 		sym = ucl_object_lookup (cur, "symbol");
@@ -1245,11 +1267,11 @@ rspamc_stat_output (FILE *out, ucl_object_t *obj)
 static void
 rspamc_output_headers (FILE *out, struct rspamd_http_message *msg)
 {
-	struct rspamd_http_header *h, *htmp;
+	struct rspamd_http_header *h;
 
-	HASH_ITER (hh, msg->headers, h, htmp) {
+	kh_foreach_value (msg->headers, h, {
 		rspamd_fprintf (out, "%T: %T\n", &h->name, &h->value);
-	}
+	});
 
 	rspamd_fprintf (out, "\n");
 }
@@ -1365,16 +1387,21 @@ rspamc_mime_output (FILE *out, ucl_object_t *result, GString *input,
 				symbuf->str,
 				0, nl_type, ",");
 		rspamd_printf_gstring (added_headers, "X-Spam-Symbols: %v%s",
-				folded_symbuf, line_end, ",");
+				folded_symbuf, line_end);
 
 		g_string_free (folded_symbuf, TRUE);
 		g_string_free (symbuf, TRUE);
 
-		if (ucl_object_lookup (result, "dkim-signature")) {
+		res = ucl_object_lookup (result, "dkim-signature");
+		if (res && res->type == UCL_STRING) {
 			rspamd_printf_gstring (added_headers, "DKIM-Signature: %s%s",
-					ucl_object_tostring (
-							ucl_object_lookup (result, "dkim-signature")),
-					line_end);
+					ucl_object_tostring (res), line_end);
+		} else if (res && res->type == UCL_ARRAY) {
+			it = NULL;
+			while ((cur = ucl_object_iterate (res, &it, true)) != NULL) {
+				rspamd_printf_gstring (added_headers, "DKIM-Signature: %s%s",
+					ucl_object_tostring (cur), line_end);
+			}
 		}
 
 		if (json || raw || compact) {
@@ -1490,6 +1517,7 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 		struct rspamd_http_message *msg,
 		const gchar *name, ucl_object_t *result, GString *input,
 		gpointer ud, gdouble start_time, gdouble send_time,
+		const gchar *body, gsize bodylen,
 		GError *err)
 {
 	gchar *ucl_out;
@@ -1497,8 +1525,6 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 	struct rspamc_command *cmd;
 	FILE *out = stdout;
 	gdouble finish = rspamd_get_ticks (FALSE), diff;
-	const gchar *body;
-	gsize body_len;
 
 	cmd = cbdata->cmd;
 
@@ -1516,17 +1542,26 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 	else {
 
 		if (cmd->cmd == RSPAMC_COMMAND_SYMBOLS && mime_output && input) {
-			rspamc_mime_output (out, result, input, diff, err);
+			if (body) {
+				GString tmp;
+
+				tmp.str = (char *)body;
+				tmp.len = bodylen;
+				rspamc_mime_output (out, result, &tmp, diff, err);
+			}
+			else {
+				rspamc_mime_output (out, result, input, diff, err);
+			}
 		}
 		else {
-			if (cmd->need_input) {
+			if (cmd->need_input && !json) {
 				if (!compact) {
 					rspamd_fprintf (out, "Results for file: %s (%.3f seconds)\n",
 							cbdata->filename, diff);
 				}
 			}
 			else {
-				if (!compact) {
+				if (!compact && !json) {
 					rspamd_fprintf (out, "Results for command: %s (%.3f seconds)\n",
 							cmd->name, diff);
 				}
@@ -1565,17 +1600,26 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 					cmd->command_output_func (out, result);
 				}
 
+				if (body) {
+					rspamd_fprintf (out, "\nNew body:\n%*s\n", (int)bodylen,
+							body);
+				}
+
 				ucl_object_unref (result);
 			}
 			else if (err != NULL) {
 				rspamd_fprintf (out, "%s\n", err->message);
 
 				if (json && msg != NULL) {
-					body = rspamd_http_message_get_body (msg, &body_len);
+					const gchar *raw;
+					gsize rawlen;
 
-					if (body) {
+					raw = rspamd_http_message_get_body (msg, &rawlen);
+
+					if (raw) {
 						/* We can also output the resulting json */
-						rspamd_fprintf (out, "%*s\n", (gint)body_len, body);
+						rspamd_fprintf (out, "%*s\n", (gint)(rawlen - bodylen),
+								raw);
 					}
 				}
 			}
@@ -1588,10 +1632,14 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 	rspamd_client_destroy (conn);
 	g_free (cbdata->filename);
 	g_free (cbdata);
+
+	if (err) {
+		retcode = EXIT_FAILURE;
+	}
 }
 
 static void
-rspamc_process_input (struct event_base *ev_base, struct rspamc_command *cmd,
+rspamc_process_input (struct ev_loop *ev_base, struct rspamc_command *cmd,
 	FILE *in, const gchar *name, GQueue *attrs)
 {
 	struct rspamd_client_connection *conn;
@@ -1649,7 +1697,7 @@ rspamc_process_input (struct event_base *ev_base, struct rspamc_command *cmd,
 
 	}
 
-	conn = rspamd_client_init (ev_base, hostbuf, port, timeout, key);
+	conn = rspamd_client_init (http_ctx, ev_base, hostbuf, port, timeout, key);
 
 	if (conn != NULL) {
 		cbdata = g_malloc0 (sizeof (struct rspamc_callback_data));
@@ -1677,7 +1725,8 @@ rspamc_process_input (struct event_base *ev_base, struct rspamc_command *cmd,
 		}
 	}
 	else {
-		rspamd_fprintf (stderr, "cannot connect to %s\n", connect_str);
+		rspamd_fprintf (stderr, "cannot connect to %s: %s\n", connect_str,
+				strerror (errno));
 		exit (EXIT_FAILURE);
 	}
 
@@ -1718,7 +1767,7 @@ rspamd_dirent_size (DIR * dirp)
 }
 
 static void
-rspamc_process_dir (struct event_base *ev_base, struct rspamc_command *cmd,
+rspamc_process_dir (struct ev_loop *ev_base, struct rspamc_command *cmd,
 	const gchar *name, GQueue *attrs)
 {
 	DIR *d;
@@ -1811,7 +1860,7 @@ rspamc_process_dir (struct event_base *ev_base, struct rspamc_command *cmd,
 				if (cur_req >= max_requests) {
 					cur_req = 0;
 					/* Wait for completion */
-					event_base_loop (ev_base, 0);
+					ev_loop (ev_base, 0);
 				}
 			}
 		}
@@ -1822,7 +1871,18 @@ rspamc_process_dir (struct event_base *ev_base, struct rspamc_command *cmd,
 	}
 
 	closedir (d);
-	event_base_loop (ev_base, 0);
+	ev_loop (ev_base, 0);
+}
+
+
+static void
+rspamc_kwattr_free (gpointer p)
+{
+	struct rspamd_http_client_header *h = (struct rspamd_http_client_header *)p;
+
+	g_free (h->value);
+	g_free (h->name);
+	g_free (h);
 }
 
 gint
@@ -1834,7 +1894,7 @@ main (gint argc, gchar **argv, gchar **env)
 	GPid cld;
 	struct rspamc_command *cmd;
 	FILE *in = NULL;
-	struct event_base *ev_base;
+	struct ev_loop *event_loop;
 	struct stat st;
 	struct sigaction sigpipe_act;
 	gchar **exclude_pattern;
@@ -1855,6 +1915,7 @@ main (gint argc, gchar **argv, gchar **env)
 	npatterns = 0;
 
 	while (exclude_pattern && *exclude_pattern) {
+		exclude_pattern ++;
 		npatterns ++;
 	}
 
@@ -1872,8 +1933,17 @@ main (gint argc, gchar **argv, gchar **env)
 		}
 	}
 
-	rspamd_init_libs ();
-	ev_base = event_base_new ();
+	struct rspamd_external_libs_ctx *libs = rspamd_init_libs ();
+	event_loop = ev_loop_new (EVBACKEND_ALL);
+
+	struct rspamd_http_context_cfg http_config;
+
+	memset (&http_config, 0, sizeof (http_config));
+	http_config.kp_cache_size_client = 32;
+	http_config.kp_cache_size_server = 0;
+	http_config.user_agent = user_agent;
+	http_ctx = rspamd_http_context_create_config (&http_config,
+			event_loop, NULL);
 
 	/* Ignore sigpipe */
 	sigemptyset (&sigpipe_act.sa_mask);
@@ -1934,10 +2004,10 @@ main (gint argc, gchar **argv, gchar **env)
 	if (start_argc == argc) {
 		/* Do command without input or with stdin */
 		if (empty_input) {
-			rspamc_process_input (ev_base, cmd, NULL, "empty", kwattrs);
+			rspamc_process_input (event_loop, cmd, NULL, "empty", kwattrs);
 		}
 		else {
-			rspamc_process_input (ev_base, cmd, in, "stdin", kwattrs);
+			rspamc_process_input (event_loop, cmd, in, "stdin", kwattrs);
 		}
 	}
 	else {
@@ -1952,7 +2022,7 @@ main (gint argc, gchar **argv, gchar **env)
 				}
 				if (S_ISDIR (st.st_mode)) {
 					/* Directories are processed with a separate limit */
-					rspamc_process_dir (ev_base, cmd, argv[i], kwattrs);
+					rspamc_process_dir (event_loop, cmd, argv[i], kwattrs);
 					cur_req = 0;
 				}
 				else {
@@ -1961,29 +2031,29 @@ main (gint argc, gchar **argv, gchar **env)
 						fprintf (stderr, "cannot open file %s\n", argv[i]);
 						exit (EXIT_FAILURE);
 					}
-					rspamc_process_input (ev_base, cmd, in, argv[i], kwattrs);
+					rspamc_process_input (event_loop, cmd, in, argv[i], kwattrs);
 					cur_req++;
 					fclose (in);
 				}
 				if (cur_req >= max_requests) {
 					cur_req = 0;
 					/* Wait for completion */
-					event_base_loop (ev_base, 0);
+					ev_loop (event_loop, 0);
 				}
 			}
 		}
 
 		if (cmd->cmd == RSPAMC_COMMAND_FUZZY_DELHASH) {
-			rspamc_process_input (ev_base, cmd, NULL, "hashes", kwattrs);
+			rspamc_process_input (event_loop, cmd, NULL, "hashes", kwattrs);
 		}
 	}
 
-	event_base_loop (ev_base, 0);
+	ev_loop (event_loop, 0);
 
-	g_queue_free_full (kwattrs, g_free);
+	g_queue_free_full (kwattrs, rspamc_kwattr_free);
 
 	/* Wait for children processes */
-	cur = g_list_first (children);
+	cur = children ? g_list_first (children) : NULL;
 	ret = 0;
 
 	while (cur) {
@@ -2017,5 +2087,8 @@ main (gint argc, gchar **argv, gchar **env)
 		g_pattern_spec_free (exclude_compiled[i]);
 	}
 
-	return ret;
+	rspamd_deinit_libs (libs);
+
+	/* Mix retcode (return from Rspamd side) and ret (return from subprocess) */
+	return ret | retcode;
 }
