@@ -13,77 +13,84 @@
  * See the License for the specific language governing permissions and
  * limitations under the License.
  */
+#include <contrib/libucl/ucl.h>
 #include "config.h"
 #include "util.h"
 #include "cfg_file.h"
 #include "rspamd.h"
 #include "message.h"
 #include "mime_expressions.h"
-#include "html.h"
+#include "libserver/html/html.h"
 #include "lua/lua_common.h"
 #include "utlist.h"
 
 gboolean rspamd_compare_encoding (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+								  GArray * args,
+								  void *unused);
 gboolean rspamd_header_exists (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+							   GArray * args,
+							   void *unused);
 gboolean rspamd_parts_distance (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+								GArray * args,
+								void *unused);
 gboolean rspamd_recipients_distance (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+									 GArray * args,
+									 void *unused);
 gboolean rspamd_has_only_html_part (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+									GArray * args,
+									void *unused);
 gboolean rspamd_is_recipients_sorted (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+									  GArray * args,
+									  void *unused);
 gboolean rspamd_compare_transfer_encoding (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+										   GArray * args,
+										   void *unused);
 gboolean rspamd_is_html_balanced (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+								  GArray * args,
+								  void *unused);
 gboolean rspamd_has_html_tag (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+							  GArray * args,
+							  void *unused);
 gboolean rspamd_has_fake_html (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+							   GArray * args,
+							   void *unused);
 static gboolean rspamd_raw_header_exists (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+										  GArray * args,
+										  void *unused);
 static gboolean rspamd_check_smtp_data (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+										GArray * args,
+										void *unused);
 static gboolean rspamd_content_type_is_type (struct rspamd_task * task,
-	GArray * args,
-	void *unused);
+											 GArray * args,
+											 void *unused);
 static gboolean rspamd_content_type_is_subtype (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+												GArray * args,
+												void *unused);
 static gboolean rspamd_content_type_has_param (struct rspamd_task * task,
-	GArray * args,
-	void *unused);
+											   GArray * args,
+											   void *unused);
 static gboolean rspamd_content_type_compare_param (struct rspamd_task * task,
-	GArray * args,
-	void *unused);
+												   GArray * args,
+												   void *unused);
 static gboolean rspamd_has_content_part (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+										 GArray * args,
+										 void *unused);
 static gboolean rspamd_has_content_part_len (struct rspamd_task *task,
-	GArray * args,
-	void *unused);
+											 GArray * args,
+											 void *unused);
 static gboolean rspamd_is_empty_body (struct rspamd_task *task,
-		GArray * args,
-		void *unused);
+									  GArray * args,
+									  void *unused);
+static gboolean rspamd_has_flag_expr (struct rspamd_task *task,
+									  GArray * args,
+									  void *unused);
+static gboolean rspamd_has_symbol_expr (struct rspamd_task *task,
+									  GArray * args,
+									  void *unused);
 
 static rspamd_expression_atom_t * rspamd_mime_expr_parse (const gchar *line, gsize len,
 		rspamd_mempool_t *pool, gpointer ud, GError **err);
-static gdouble rspamd_mime_expr_process (struct rspamd_expr_process_data *process_data, rspamd_expression_atom_t *atom);
+static gdouble rspamd_mime_expr_process (void *ud, rspamd_expression_atom_t *atom);
 static gint rspamd_mime_expr_priority (rspamd_expression_atom_t *atom);
 static void rspamd_mime_expr_destroy (rspamd_expression_atom_t *atom);
 
@@ -114,7 +121,8 @@ struct rspamd_function_atom {
 enum rspamd_mime_atom_type {
 	MIME_ATOM_REGEXP = 0,
 	MIME_ATOM_INTERNAL_FUNCTION,
-	MIME_ATOM_LUA_FUNCTION
+	MIME_ATOM_LUA_FUNCTION,
+	MIME_ATOM_LOCAL_LUA_FUNCTION, /* New style */
 };
 
 struct rspamd_mime_atom {
@@ -123,6 +131,7 @@ struct rspamd_mime_atom {
 		struct rspamd_regexp_atom *re;
 		struct rspamd_function_atom *func;
 		const gchar *lua_function;
+		gint lua_cbref;
 	} d;
 	enum rspamd_mime_atom_type type;
 };
@@ -136,25 +145,27 @@ static struct _fl {
 	rspamd_internal_func_t func;
 	void *user_data;
 } rspamd_functions_list[] = {
-	{"check_smtp_data", rspamd_check_smtp_data, NULL},
-	{"compare_encoding", rspamd_compare_encoding, NULL},
-	{"compare_parts_distance", rspamd_parts_distance, NULL},
-	{"compare_recipients_distance", rspamd_recipients_distance, NULL},
-	{"compare_transfer_encoding", rspamd_compare_transfer_encoding, NULL},
-	{"content_type_compare_param", rspamd_content_type_compare_param, NULL},
-	{"content_type_has_param", rspamd_content_type_has_param, NULL},
-	{"content_type_is_subtype", rspamd_content_type_is_subtype, NULL},
-	{"content_type_is_type", rspamd_content_type_is_type, NULL},
-	{"has_content_part", rspamd_has_content_part, NULL},
-	{"has_content_part_len", rspamd_has_content_part_len, NULL},
-	{"has_fake_html", rspamd_has_fake_html, NULL},
-	{"has_html_tag", rspamd_has_html_tag, NULL},
-	{"has_only_html_part", rspamd_has_only_html_part, NULL},
-	{"header_exists", rspamd_header_exists, NULL},
-	{"is_empty_body", rspamd_is_empty_body, NULL},
-	{"is_html_balanced", rspamd_is_html_balanced, NULL},
-	{"is_recipients_sorted", rspamd_is_recipients_sorted, NULL},
-	{"raw_header_exists", rspamd_raw_header_exists, NULL}
+		{"check_smtp_data", rspamd_check_smtp_data, NULL},
+		{"compare_encoding", rspamd_compare_encoding, NULL},
+		{"compare_parts_distance", rspamd_parts_distance, NULL},
+		{"compare_recipients_distance", rspamd_recipients_distance, NULL},
+		{"compare_transfer_encoding", rspamd_compare_transfer_encoding, NULL},
+		{"content_type_compare_param", rspamd_content_type_compare_param, NULL},
+		{"content_type_has_param", rspamd_content_type_has_param, NULL},
+		{"content_type_is_subtype", rspamd_content_type_is_subtype, NULL},
+		{"content_type_is_type", rspamd_content_type_is_type, NULL},
+		{"has_content_part", rspamd_has_content_part, NULL},
+		{"has_content_part_len", rspamd_has_content_part_len, NULL},
+		{"has_fake_html", rspamd_has_fake_html, NULL},
+		{"has_flag", rspamd_has_flag_expr, NULL},
+		{"has_html_tag", rspamd_has_html_tag, NULL},
+		{"has_only_html_part", rspamd_has_only_html_part, NULL},
+		{"has_symbol", rspamd_has_symbol_expr, NULL},
+		{"header_exists", rspamd_header_exists, NULL},
+		{"is_empty_body", rspamd_is_empty_body, NULL},
+		{"is_html_balanced", rspamd_is_html_balanced, NULL},
+		{"is_recipients_sorted", rspamd_is_recipients_sorted, NULL},
+		{"raw_header_exists", rspamd_raw_header_exists, NULL},
 };
 
 const struct rspamd_atom_subr mime_expr_subr = {
@@ -230,6 +241,10 @@ rspamd_parse_long_option (const gchar *start, gsize len,
 		ret = TRUE;
 		a->type = RSPAMD_RE_URL;
 	}
+	else if (TYPE_CHECK (start, "email", len)) {
+		ret = TRUE;
+		a->type = RSPAMD_RE_EMAIL;
+	}
 	else if (TYPE_CHECK (start, "sa_body", len)) {
 		ret = TRUE;
 		a->type = RSPAMD_RE_SABODY;
@@ -239,6 +254,18 @@ rspamd_parse_long_option (const gchar *start, gsize len,
 		ret = TRUE;
 		a->type = RSPAMD_RE_SARAWBODY;
 	}
+	else if (TYPE_CHECK (start, "words", len)) {
+		ret = TRUE;
+		a->type = RSPAMD_RE_WORDS;
+	}
+	else if (TYPE_CHECK (start, "raw_words", len)) {
+		ret = TRUE;
+		a->type = RSPAMD_RE_RAWWORDS;
+	}
+	else if (TYPE_CHECK (start, "stem_words", len)) {
+		ret = TRUE;
+		a->type = RSPAMD_RE_STEMWORDS;
+	}
 	else if (TYPE_CHECK (start, "selector", len)) {
 		ret = TRUE;
 		a->type = RSPAMD_RE_SELECTOR;
@@ -345,6 +372,8 @@ rspamd_mime_expr_parse_regexp_atom (rspamd_mempool_t * pool, const gchar *line,
 		case 'u':
 		case 'O':
 		case 'r':
+		case 'L':
+			/* Handled by rspamd_regexp_t */
 			g_string_append_c (re_flags, *p);
 			p++;
 			break;
@@ -357,7 +386,7 @@ rspamd_mime_expr_parse_regexp_atom (rspamd_mempool_t * pool, const gchar *line,
 			p++;
 			break;
 		case 'R':
-			result->type = RSPAMD_RE_RAWHEADER;
+			result->type = RSPAMD_RE_ALLHEADER;
 			p++;
 			break;
 		case 'B':
@@ -621,8 +650,9 @@ rspamd_mime_expr_parse (const gchar *line, gsize len,
 {
 	rspamd_expression_atom_t *a = NULL;
 	struct rspamd_mime_atom *mime_atom = NULL;
-	const gchar *p, *end;
-	struct rspamd_config *cfg = ud;
+	const gchar *p, *end, *c = NULL;
+	struct rspamd_mime_expr_ud *real_ud = (struct rspamd_mime_expr_ud *)ud;
+	struct rspamd_config *cfg;
 	rspamd_regexp_t *own_re;
 	gchar t;
 	gint type = MIME_ATOM_REGEXP, obraces = 0, ebraces = 0;
@@ -636,6 +666,7 @@ rspamd_mime_expr_parse (const gchar *line, gsize len,
 		in_flags_brace,
 		got_obrace,
 		in_function,
+		in_local_function,
 		got_ebrace,
 		end_atom,
 		bad_atom
@@ -643,6 +674,7 @@ rspamd_mime_expr_parse (const gchar *line, gsize len,
 
 	p = line;
 	end = p + len;
+	cfg = real_ud->cfg;
 
 	while (p < end) {
 		t = *p;
@@ -658,11 +690,20 @@ rspamd_mime_expr_parse (const gchar *line, gsize len,
 				state = got_obrace;
 			}
 			else if (!g_ascii_isalnum (t) && t != '_' && t != '-' && t != '=') {
-				/* Likely lua function, identified by just a string */
-				type = MIME_ATOM_LUA_FUNCTION;
-				state = end_atom;
-				/* Do not increase p */
-				continue;
+				if (t == ':') {
+					if (p - line == 3 && memcmp (line, "lua", 3) == 0) {
+						type = MIME_ATOM_LOCAL_LUA_FUNCTION;
+						state = in_local_function;
+						c = p + 1;
+					}
+				}
+				else {
+					/* Likely lua function, identified by just a string */
+					type = MIME_ATOM_LUA_FUNCTION;
+					state = end_atom;
+					/* Do not increase p */
+					continue;
+				}
 			}
 			else if (g_ascii_isspace (t)) {
 				state = bad_atom;
@@ -728,6 +769,15 @@ rspamd_mime_expr_parse (const gchar *line, gsize len,
 			}
 			p ++;
 			break;
+		case in_local_function:
+			if (!(g_ascii_isalnum (t) || t == '-' || t == '_')) {
+				g_assert (c != NULL);
+				state = end_atom;
+			}
+			else {
+				p++;
+			}
+			break;
 		case got_ebrace:
 			state = end_atom;
 			break;
@@ -759,11 +809,49 @@ set:
 		mime_atom->d.re = rspamd_mime_expr_parse_regexp_atom (pool,
 				mime_atom->str, cfg);
 		if (mime_atom->d.re == NULL) {
-			g_set_error (err, rspamd_mime_expr_quark(), 200, "cannot parse regexp '%s'",
+			g_set_error (err, rspamd_mime_expr_quark(), 200,
+					"cannot parse regexp '%s'",
 					mime_atom->str);
 			goto err;
 		}
 		else {
+			gint lua_cbref = -1;
+
+			/* Check regexp condition */
+			if (real_ud->conf_obj != NULL) {
+				const ucl_object_t *re_conditions = ucl_object_lookup (real_ud->conf_obj,
+						"re_conditions");
+
+				if (re_conditions != NULL) {
+					if (ucl_object_type (re_conditions) != UCL_OBJECT) {
+						g_set_error (err, rspamd_mime_expr_quark (), 320,
+								"re_conditions is not a table for '%s'",
+								mime_atom->str);
+						goto err;
+					}
+
+					const ucl_object_t *function_obj = ucl_object_lookup (re_conditions,
+							mime_atom->str);
+
+					if (function_obj != NULL) {
+						if (ucl_object_type (function_obj) != UCL_USERDATA) {
+							g_set_error (err, rspamd_mime_expr_quark (), 320,
+									"condition for '%s' is invalid, must be function",
+									mime_atom->str);
+							goto err;
+						}
+
+						struct ucl_lua_funcdata *fd = function_obj->value.ud;
+
+						lua_cbref = fd->idx;
+					}
+				}
+			}
+
+			if (lua_cbref != -1) {
+				msg_info_config ("added condition for regexp %s", mime_atom->str);
+			}
+
 			/* Register new item in the cache */
 			if (mime_atom->d.re->type == RSPAMD_RE_HEADER ||
 					mime_atom->d.re->type == RSPAMD_RE_RAWHEADER ||
@@ -775,7 +863,8 @@ set:
 							mime_atom->d.re->regexp,
 							mime_atom->d.re->type,
 							mime_atom->d.re->extra.header,
-							strlen (mime_atom->d.re->extra.header) + 1);
+							strlen (mime_atom->d.re->extra.header) + 1,
+							lua_cbref);
 					/* Pass ownership to the cache */
 					rspamd_regexp_unref (own_re);
 				}
@@ -786,6 +875,7 @@ set:
 							200,
 							"no header name in header regexp: '%s'",
 							mime_atom->str);
+					rspamd_regexp_unref (mime_atom->d.re->regexp);
 					goto err;
 				}
 
@@ -797,17 +887,19 @@ set:
 							mime_atom->d.re->regexp,
 							mime_atom->d.re->type,
 							mime_atom->d.re->extra.selector,
-							strlen (mime_atom->d.re->extra.selector) + 1);
+							strlen (mime_atom->d.re->extra.selector) + 1,
+							lua_cbref);
 					/* Pass ownership to the cache */
 					rspamd_regexp_unref (own_re);
 				}
 				else {
-					/* We have header regexp, but no header name is detected */
+					/* We have selector regexp, but no selector name is detected */
 					g_set_error (err,
 							rspamd_mime_expr_quark (),
 							200,
 							"no selector name in selector regexp: '%s'",
 							mime_atom->str);
+					rspamd_regexp_unref (mime_atom->d.re->regexp);
 					goto err;
 				}
 			}
@@ -817,7 +909,8 @@ set:
 						mime_atom->d.re->regexp,
 						mime_atom->d.re->type,
 						NULL,
-						0);
+						0,
+						lua_cbref);
 				/* Pass ownership to the cache */
 				rspamd_regexp_unref (own_re);
 			}
@@ -829,19 +922,72 @@ set:
 		lua_getglobal (cfg->lua_state, mime_atom->str);
 
 		if (lua_type (cfg->lua_state, -1) != LUA_TFUNCTION) {
-			g_set_error (err, rspamd_mime_expr_quark(), 200, "no such lua function '%s'",
+			g_set_error (err, rspamd_mime_expr_quark(), 200,
+					"no such lua function '%s'",
 					mime_atom->str);
 			lua_pop (cfg->lua_state, 1);
 
 			goto err;
 		}
+
 		lua_pop (cfg->lua_state, 1);
 	}
+	else if (type == MIME_ATOM_LOCAL_LUA_FUNCTION) {
+		/* p pointer is set to the start of Lua function name */
+
+		if (real_ud->conf_obj == NULL) {
+			g_set_error (err, rspamd_mime_expr_quark(), 300,
+					"no config object for '%s'",
+					mime_atom->str);
+			goto err;
+		}
+
+		const ucl_object_t *functions = ucl_object_lookup (real_ud->conf_obj,
+				"functions");
+
+		if (functions == NULL) {
+			g_set_error (err, rspamd_mime_expr_quark(), 310,
+					"no functions defined for '%s'",
+					mime_atom->str);
+			goto err;
+		}
+
+		if (ucl_object_type (functions) != UCL_OBJECT) {
+			g_set_error (err, rspamd_mime_expr_quark(), 320,
+					"functions is not a table for '%s'",
+					mime_atom->str);
+			goto err;
+		}
+
+		const ucl_object_t *function_obj;
+
+		function_obj = ucl_object_lookup_len (functions, c,
+				p - c);
+
+		if (function_obj == NULL) {
+			g_set_error (err, rspamd_mime_expr_quark(), 320,
+					"function %*.s is not found for '%s'",
+					(int)(p - c), c, mime_atom->str);
+			goto err;
+		}
+
+		if (ucl_object_type (function_obj) != UCL_USERDATA) {
+			g_set_error (err, rspamd_mime_expr_quark(), 320,
+					"function %*.s has invalid type for '%s'",
+					(int)(p - c), c, mime_atom->str);
+			goto err;
+		}
+
+		struct ucl_lua_funcdata *fd = function_obj->value.ud;
+
+		mime_atom->d.lua_cbref = fd->idx;
+	}
 	else {
 		mime_atom->d.func = rspamd_mime_expr_parse_function_atom (pool,
 				mime_atom->str);
 		if (mime_atom->d.func == NULL) {
-			g_set_error (err, rspamd_mime_expr_quark(), 200, "cannot parse function '%s'",
+			g_set_error (err, rspamd_mime_expr_quark(), 200,
+					"cannot parse function '%s'",
 					mime_atom->str);
 			goto err;
 		}
@@ -914,26 +1060,34 @@ rspamd_mime_expr_priority (rspamd_expression_atom_t *atom)
 	switch (mime_atom->type) {
 	case MIME_ATOM_INTERNAL_FUNCTION:
 		/* Prioritize internal functions slightly */
-		ret = 50;
+		ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 8;
 		break;
 	case MIME_ATOM_LUA_FUNCTION:
-		ret = 50;
+	case MIME_ATOM_LOCAL_LUA_FUNCTION:
+		ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 4;
 		break;
 	case MIME_ATOM_REGEXP:
 		switch (mime_atom->d.re->type) {
 		case RSPAMD_RE_HEADER:
 		case RSPAMD_RE_RAWHEADER:
-			ret = 100;
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 16;
 			break;
 		case RSPAMD_RE_URL:
-			ret = 90;
+		case RSPAMD_RE_EMAIL:
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 8;
+			break;
+		case RSPAMD_RE_SELECTOR:
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 8;
 			break;
 		case RSPAMD_RE_MIME:
 		case RSPAMD_RE_RAWMIME:
-			ret = 10;
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 2;
 			break;
+		case RSPAMD_RE_WORDS:
+		case RSPAMD_RE_RAWWORDS:
+		case RSPAMD_RE_STEMWORDS:
 		default:
-			/* For message regexp */
+			/* For expensive regexps */
 			ret = 0;
 			break;
 		}
@@ -978,9 +1132,9 @@ rspamd_mime_expr_process_function (struct rspamd_function_atom * func,
 }
 
 static gdouble
-rspamd_mime_expr_process (struct rspamd_expr_process_data *process_data, rspamd_expression_atom_t *atom)
+rspamd_mime_expr_process (void *ud, rspamd_expression_atom_t *atom)
 {
-	struct rspamd_task *task = process_data->task;
+	struct rspamd_task *task = (struct rspamd_task *)ud;
 	struct rspamd_mime_atom *mime_atom;
 	lua_State *L;
 	gdouble ret = 0;
@@ -1020,6 +1174,36 @@ rspamd_mime_expr_process (struct rspamd_expr_process_data *process_data, rspamd_
 			lua_pop (L, 1);
 		}
 	}
+	else if (mime_atom->type == MIME_ATOM_LOCAL_LUA_FUNCTION) {
+		gint err_idx;
+
+		L = task->cfg->lua_state;
+		lua_pushcfunction (L, &rspamd_lua_traceback);
+		err_idx = lua_gettop (L);
+
+		lua_rawgeti (L, LUA_REGISTRYINDEX, mime_atom->d.lua_cbref);
+		rspamd_lua_task_push (L, task);
+
+		if (lua_pcall (L, 1, 1, err_idx) != 0) {
+			msg_info_task ("lua call to local function for atom '%s' failed: %s",
+					mime_atom->str,
+					lua_tostring (L, -1));
+		}
+		else {
+			if (lua_type (L, -1) == LUA_TBOOLEAN) {
+				ret = lua_toboolean (L, -1);
+			}
+			else if (lua_type (L, -1) == LUA_TNUMBER) {
+				ret = lua_tonumber (L, 1);
+			}
+			else {
+				msg_err_task ("%s returned wrong return type: %s",
+						mime_atom->str, lua_typename (L, lua_type (L, -1)));
+			}
+		}
+
+		lua_settop (L, 0);
+	}
 	else {
 		ret = rspamd_mime_expr_process_function (mime_atom->d.func, task,
 				task->cfg->lua_state);
@@ -1074,7 +1258,7 @@ gboolean
 rspamd_header_exists (struct rspamd_task * task, GArray * args, void *unused)
 {
 	struct expression_argument *arg;
-	GPtrArray *headerlist;
+	struct rspamd_mime_header *rh;
 
 	if (args == NULL || task == NULL) {
 		return FALSE;
@@ -1086,14 +1270,13 @@ rspamd_header_exists (struct rspamd_task * task, GArray * args, void *unused)
 		return FALSE;
 	}
 
-	headerlist = rspamd_message_get_header_array (task,
-			(gchar *)arg->data,
-			FALSE);
+	rh = rspamd_message_get_header_array(task,
+			(gchar *) arg->data, FALSE);
 
 	debug_task ("try to get header %s: %d", (gchar *)arg->data,
-			(headerlist != NULL));
+			(rh != NULL));
 
-	if (headerlist) {
+	if (rh) {
 		return TRUE;
 	}
 
@@ -1184,6 +1367,19 @@ struct addr_list {
 	guint addrlen;
 };
 
+static gint
+addr_list_cmp_func (const void *a, const void *b)
+{
+	const struct addr_list *addra = (struct addr_list *)a,
+			*addrb = (struct addr_list *)b;
+
+	if (addra->addrlen != addrb->addrlen) {
+		return addra->addrlen - addrb->addrlen;
+	}
+
+	return memcmp (addra->addr, addrb->addr, addra->addrlen);
+}
+
 #define COMPARE_RCPT_LEN 3
 #define MIN_RCPT_TO_COMPARE 7
 
@@ -1195,7 +1391,7 @@ rspamd_recipients_distance (struct rspamd_task *task, GArray * args,
 	struct rspamd_email_address *cur;
 	double threshold;
 	struct addr_list *ar;
-	gint num, i, j, hits = 0, total = 0;
+	gint num, i, hits = 0;
 
 	if (args == NULL) {
 		msg_warn_task ("no parameters to function");
@@ -1218,11 +1414,11 @@ rspamd_recipients_distance (struct rspamd_task *task, GArray * args,
 		return FALSE;
 	}
 
-	if (!task->rcpt_mime) {
+	if (!MESSAGE_FIELD (task, rcpt_mime)) {
 		return FALSE;
 	}
 
-	num = task->rcpt_mime->len;
+	num = MESSAGE_FIELD (task, rcpt_mime)->len;
 
 	if (num < MIN_RCPT_TO_COMPARE) {
 		return FALSE;
@@ -1231,27 +1427,31 @@ rspamd_recipients_distance (struct rspamd_task *task, GArray * args,
 	ar = rspamd_mempool_alloc0 (task->task_pool, num * sizeof (struct addr_list));
 
 	/* Fill array */
-	PTR_ARRAY_FOREACH (task->rcpt_mime, i, cur) {
-		ar[i].name = cur->addr;
-		ar[i].namelen = cur->addr_len;
-		ar[i].addr = cur->domain;
-		ar[i].addrlen = cur->domain_len;
-	}
-
-	/* Cycle all elements in array */
-	for (i = 0; i < num; i++) {
-		for (j = i + 1; j < num; j++) {
-			if (ar[i].namelen >= COMPARE_RCPT_LEN && ar[j].namelen >= COMPARE_RCPT_LEN &&
-				rspamd_lc_cmp (ar[i].name, ar[j].name, COMPARE_RCPT_LEN) == 0) {
-				/* Common name part */
-				hits++;
-			}
-
-			total++;
+	num = 0;
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, rcpt_mime), i, cur) {
+		if (cur->addr_len > COMPARE_RCPT_LEN) {
+			ar[num].name = cur->addr;
+			ar[num].namelen = cur->addr_len;
+			ar[num].addr = cur->domain;
+			ar[num].addrlen = cur->domain_len;
+			num ++;
 		}
 	}
 
-	if ((hits * num / 2.) / (double)total >= threshold) {
+	qsort (ar, num, sizeof (*ar), addr_list_cmp_func);
+
+	/* Cycle all elements in array */
+	for (i = 0; i < num; i++) {
+		if (i < num - 1) {
+			if (ar[i].namelen == ar[i + 1].namelen) {
+				if (rspamd_lc_cmp (ar[i].name, ar[i + 1].name, COMPARE_RCPT_LEN) == 0) {
+					hits++;
+				}
+			}
+		}
+	}
+
+	if ((hits * num / 2.) / (double)num >= threshold) {
 		return TRUE;
 	}
 
@@ -1263,20 +1463,22 @@ rspamd_has_only_html_part (struct rspamd_task * task, GArray * args,
 	void *unused)
 {
 	struct rspamd_mime_text_part *p;
-	gboolean res = FALSE;
+	guint i, cnt_html = 0, cnt_txt = 0;
 
-	if (task->text_parts->len == 1) {
-		p = g_ptr_array_index (task->text_parts, 0);
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, text_parts), i, p) {
+		p = g_ptr_array_index (MESSAGE_FIELD (task, text_parts), 0);
 
-		if (IS_PART_HTML (p)) {
-			res = TRUE;
-		}
-		else {
-			res = FALSE;
+		if (!IS_TEXT_PART_ATTACHMENT (p)) {
+			if (IS_TEXT_PART_HTML (p)) {
+				cnt_html++;
+			}
+			else {
+				cnt_txt++;
+			}
 		}
 	}
 
-	return res;
+	return (cnt_html > 0 && cnt_txt == 0);
 }
 
 static gboolean
@@ -1319,8 +1521,8 @@ rspamd_is_recipients_sorted (struct rspamd_task * task,
 {
 	/* Check all types of addresses */
 
-	if (task->rcpt_mime) {
-		return is_recipient_list_sorted (task->rcpt_mime);
+	if (MESSAGE_FIELD (task, rcpt_mime)) {
+		return is_recipient_list_sorted (MESSAGE_FIELD (task, rcpt_mime));
 	}
 
 	return FALSE;
@@ -1354,8 +1556,8 @@ rspamd_compare_transfer_encoding (struct rspamd_task * task,
 		return FALSE;
 	}
 
-	PTR_ARRAY_FOREACH (task->parts, i, part) {
-		if (IS_CT_TEXT (part->ct)) {
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, part) {
+		if (IS_PART_TEXT (part)) {
 			if (part->cte == cte) {
 				return TRUE;
 			}
@@ -1368,26 +1570,8 @@ rspamd_compare_transfer_encoding (struct rspamd_task * task,
 gboolean
 rspamd_is_html_balanced (struct rspamd_task * task, GArray * args, void *unused)
 {
-	struct rspamd_mime_text_part *p;
-	guint i;
-	gboolean res = TRUE;
-
-	for (i = 0; i < task->text_parts->len; i ++) {
-
-		p = g_ptr_array_index (task->text_parts, i);
-		if (!IS_PART_EMPTY (p) && IS_PART_HTML (p)) {
-			if (p->flags & RSPAMD_MIME_TEXT_PART_FLAG_BALANCED) {
-				res = TRUE;
-			}
-			else {
-				res = FALSE;
-				break;
-			}
-		}
-	}
-
-	return res;
-
+	/* Totally broken but seems to be never used */
+	return TRUE;
 }
 
 gboolean
@@ -1409,10 +1593,8 @@ rspamd_has_html_tag (struct rspamd_task * task, GArray * args, void *unused)
 		return FALSE;
 	}
 
-	for (i = 0; i < task->text_parts->len; i ++) {
-		p = g_ptr_array_index (task->text_parts, i);
-
-		if (!IS_PART_EMPTY (p) && IS_PART_HTML (p) && p->html) {
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, text_parts), i, p) {
+		if (IS_TEXT_PART_HTML (p) && p->html) {
 			res = rspamd_html_tag_seen (p->html, arg->data);
 		}
 
@@ -1432,10 +1614,8 @@ rspamd_has_fake_html (struct rspamd_task * task, GArray * args, void *unused)
 	guint i;
 	gboolean res = FALSE;
 
-	for (i = 0; i < task->text_parts->len; i ++) {
-		p = g_ptr_array_index (task->text_parts, i);
-
-		if (!IS_PART_EMPTY (p) && IS_PART_HTML (p) && p->html->html_tags == NULL) {
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, text_parts), i, p) {
+		if (IS_TEXT_PART_HTML (p) && (rspamd_html_get_tags_count(p->html) < 2)) {
 			res = TRUE;
 		}
 
@@ -1463,7 +1643,7 @@ rspamd_raw_header_exists (struct rspamd_task *task, GArray * args, void *unused)
 		return FALSE;
 	}
 
-	return g_hash_table_lookup (task->raw_headers, arg->data) != NULL;
+	return rspamd_message_get_header_array(task, arg->data, FALSE) != NULL;
 }
 
 static gboolean
@@ -1553,7 +1733,7 @@ rspamd_check_smtp_data (struct rspamd_task *task, GArray * args, void *unused)
 		case 's':
 		case 'S':
 			if (g_ascii_strcasecmp (type, "subject") == 0) {
-				str = task->subject;
+				str = MESSAGE_FIELD (task, subject);
 			}
 			else {
 				msg_warn_task ("bad argument to function: %s", type);
@@ -1611,12 +1791,45 @@ rspamd_check_smtp_data (struct rspamd_task *task, GArray * args, void *unused)
 	return FALSE;
 }
 
+static inline gboolean
+rspamd_check_ct_attr (const gchar *begin, gsize len,
+		struct expression_argument *arg_pattern)
+{
+	rspamd_regexp_t *re;
+	gboolean r = FALSE;
+
+	if (arg_pattern->type == EXPRESSION_ARGUMENT_REGEXP) {
+		re = arg_pattern->data;
+
+		if (len > 0) {
+			r = rspamd_regexp_search (re,
+					begin, len,
+					NULL, NULL, FALSE, NULL);
+		}
+
+		if (r) {
+			return TRUE;
+		}
+	}
+	else {
+		/* Just do strcasecmp */
+		gsize plen = strlen (arg_pattern->data);
+
+		if (plen == len &&
+			g_ascii_strncasecmp (arg_pattern->data, begin, len) == 0) {
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 static gboolean
 rspamd_content_type_compare_param (struct rspamd_task * task,
 	GArray * args,
 	void *unused)
 {
-	rspamd_regexp_t *re;
+
 	struct expression_argument *arg, *arg1, *arg_pattern;
 	gboolean recursive = FALSE;
 	struct rspamd_mime_part *cur_part;
@@ -1624,7 +1837,6 @@ rspamd_content_type_compare_param (struct rspamd_task * task,
 	rspamd_ftok_t srch;
 	struct rspamd_content_type_param *found = NULL, *cur;
 	const gchar *param_name;
-	gboolean r = FALSE;
 
 	if (args == NULL || args->len < 2) {
 		msg_warn_task ("no parameters to function");
@@ -1636,9 +1848,7 @@ rspamd_content_type_compare_param (struct rspamd_task * task,
 	param_name = arg->data;
 	arg_pattern = &g_array_index (args, struct expression_argument, 1);
 
-	for (i = 0; i < task->parts->len; i ++) {
-		cur_part = g_ptr_array_index (task->parts, i);
-
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, cur_part) {
 		if (args->len >= 3) {
 			arg1 = &g_array_index (args, struct expression_argument, 2);
 			if (g_ascii_strncasecmp (arg1->data, "true",
@@ -1651,37 +1861,38 @@ rspamd_content_type_compare_param (struct rspamd_task * task,
 			 * If user did not specify argument, let's assume that he wants
 			 * recursive search if mime part is multipart/mixed
 			 */
-			if (IS_CT_MULTIPART (cur_part->ct)) {
+			if (IS_PART_MULTIPART (cur_part)) {
 				recursive = TRUE;
 			}
 		}
 
-		if (cur_part->ct->attrs) {
-			RSPAMD_FTOK_FROM_STR (&srch, param_name);
+		rspamd_ftok_t lit;
+		RSPAMD_FTOK_FROM_STR (&srch, param_name);
+		RSPAMD_FTOK_FROM_STR (&lit, "charset");
 
+		if (rspamd_ftok_equal (&srch, &lit)) {
+			if (rspamd_check_ct_attr (cur_part->ct->charset.begin,
+					cur_part->ct->charset.len, arg_pattern)) {
+				return TRUE;
+			}
+		}
+
+		RSPAMD_FTOK_FROM_STR (&lit, "boundary");
+		if (rspamd_ftok_equal (&srch, &lit)) {
+			if (rspamd_check_ct_attr (cur_part->ct->orig_boundary.begin,
+					cur_part->ct->orig_boundary.len, arg_pattern)) {
+				return TRUE;
+			}
+		}
+
+		if (cur_part->ct->attrs) {
 			found = g_hash_table_lookup (cur_part->ct->attrs, &srch);
 
 			if (found) {
 				DL_FOREACH (found, cur) {
-					if (arg_pattern->type == EXPRESSION_ARGUMENT_REGEXP) {
-						re = arg_pattern->data;
-
-						if (cur->value.len > 0) {
-							r = rspamd_regexp_search (re,
-									cur->value.begin, cur->value.len,
-									NULL, NULL, FALSE, NULL);
-						}
-
-						if (r) {
-							return TRUE;
-						}
-					}
-					else {
-						/* Just do strcasecmp */
-						RSPAMD_FTOK_FROM_STR (&srch, arg_pattern->data);
-						if (rspamd_ftok_casecmp (&srch, &cur->value) == 0) {
-							return TRUE;
-						}
+					if (rspamd_check_ct_attr (cur->value.begin,
+							cur->value.len, arg_pattern)) {
+						return TRUE;
 					}
 				}
 			}
@@ -1717,9 +1928,7 @@ rspamd_content_type_has_param (struct rspamd_task * task,
 	g_assert (arg->type == EXPRESSION_ARGUMENT_NORMAL);
 	param_name = arg->data;
 
-	for (i = 0; i < task->parts->len; i ++) {
-		cur_part = g_ptr_array_index (task->parts, i);
-
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, cur_part) {
 		if (args->len >= 2) {
 			arg1 = &g_array_index (args, struct expression_argument, 1);
 			if (g_ascii_strncasecmp (arg1->data, "true",
@@ -1732,14 +1941,30 @@ rspamd_content_type_has_param (struct rspamd_task * task,
 			 * If user did not specify argument, let's assume that he wants
 			 * recursive search if mime part is multipart/mixed
 			 */
-			if (IS_CT_MULTIPART (cur_part->ct)) {
+			if (IS_PART_MULTIPART (cur_part)) {
 				recursive = TRUE;
 			}
 		}
 
-		if (cur_part->ct->attrs) {
-			RSPAMD_FTOK_FROM_STR (&srch, param_name);
 
+		rspamd_ftok_t lit;
+		RSPAMD_FTOK_FROM_STR (&srch, param_name);
+		RSPAMD_FTOK_FROM_STR (&lit, "charset");
+
+		if (rspamd_ftok_equal (&srch, &lit)) {
+			if (cur_part->ct->charset.len > 0) {
+				return TRUE;
+			}
+		}
+
+		RSPAMD_FTOK_FROM_STR (&lit, "boundary");
+		if (rspamd_ftok_equal (&srch, &lit)) {
+			if (cur_part->ct->boundary.len > 0) {
+				return TRUE;
+			}
+		}
+
+		if (cur_part->ct->attrs) {
 			found = g_hash_table_lookup (cur_part->ct->attrs, &srch);
 
 			if (found) {
@@ -1776,8 +2001,7 @@ rspamd_content_type_check (struct rspamd_task *task,
 
 	arg_pattern = &g_array_index (args, struct expression_argument, 0);
 
-	for (i = 0; i < task->parts->len; i ++) {
-		cur_part = g_ptr_array_index (task->parts, i);
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, cur_part) {
 		ct = cur_part->ct;
 
 		if (args->len >= 2) {
@@ -1792,7 +2016,7 @@ rspamd_content_type_check (struct rspamd_task *task,
 			 * If user did not specify argument, let's assume that he wants
 			 * recursive search if mime part is multipart/mixed
 			 */
-			if (IS_CT_MULTIPART (ct)) {
+			if (IS_PART_MULTIPART (cur_part)) {
 				recursive = TRUE;
 			}
 		}
@@ -1916,8 +2140,7 @@ common_has_content_part (struct rspamd_task * task,
 	gint r = 0;
 	guint i;
 
-	for (i = 0; i < task->parts->len; i ++) {
-		part = g_ptr_array_index (task->parts, i);
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, part) {
 		ct = part->ct;
 
 		if (ct == NULL) {
@@ -2041,11 +2264,112 @@ rspamd_is_empty_body (struct rspamd_task *task,
 	struct rspamd_mime_part *part;
 	guint i;
 
-	PTR_ARRAY_FOREACH (task->parts, i, part) {
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, part) {
 		if (part->parsed_data.len > 0) {
 			return FALSE;
 		}
 	}
 
 	return TRUE;
-}
\ No newline at end of file
+}
+
+#define TASK_FLAG_READ(flag) do { \
+	result = !!(task->flags & (flag)); \
+} while(0)
+
+#define TASK_GET_FLAG(flag, strname, macro) do { \
+	if (!found && strcmp ((flag), strname) == 0) { \
+		TASK_FLAG_READ((macro)); \
+		found = TRUE; \
+	} \
+} while(0)
+
+#define TASK_PROTOCOL_FLAG_READ(flag) do { \
+	result = !!(task->protocol_flags & (flag)); \
+} while(0)
+
+#define TASK_GET_PROTOCOL_FLAG(flag, strname, macro) do { \
+	if (!found && strcmp ((flag), strname) == 0) { \
+		TASK_PROTOCOL_FLAG_READ((macro)); \
+		found = TRUE; \
+	} \
+} while(0)
+
+
+static gboolean
+rspamd_has_flag_expr (struct rspamd_task *task,
+					  GArray * args,
+					  void *unused)
+{
+	gboolean found = FALSE, result = FALSE;
+	struct expression_argument *flag_arg;
+	const gchar *flag_str;
+
+	if (args == NULL) {
+		msg_warn_task ("no parameters to function");
+		return FALSE;
+	}
+
+	flag_arg = &g_array_index (args, struct expression_argument, 0);
+
+	if (flag_arg->type != EXPRESSION_ARGUMENT_NORMAL) {
+		msg_warn_task ("invalid parameter to function");
+		return FALSE;
+	}
+
+	flag_str = (const gchar *)flag_arg->data;
+
+	TASK_GET_FLAG (flag_str, "pass_all", RSPAMD_TASK_FLAG_PASS_ALL);
+	TASK_GET_FLAG (flag_str, "no_log", RSPAMD_TASK_FLAG_NO_LOG);
+	TASK_GET_FLAG (flag_str, "no_stat", RSPAMD_TASK_FLAG_NO_STAT);
+	TASK_GET_FLAG (flag_str, "skip", RSPAMD_TASK_FLAG_SKIP);
+	TASK_GET_PROTOCOL_FLAG (flag_str, "extended_urls",
+			RSPAMD_TASK_PROTOCOL_FLAG_EXT_URLS);
+	TASK_GET_FLAG (flag_str, "learn_spam", RSPAMD_TASK_FLAG_LEARN_SPAM);
+	TASK_GET_FLAG (flag_str, "learn_ham", RSPAMD_TASK_FLAG_LEARN_HAM);
+	TASK_GET_FLAG (flag_str, "greylisted", RSPAMD_TASK_FLAG_GREYLISTED);
+	TASK_GET_FLAG (flag_str, "broken_headers",
+			RSPAMD_TASK_FLAG_BROKEN_HEADERS);
+	TASK_GET_FLAG (flag_str, "skip_process",
+			RSPAMD_TASK_FLAG_SKIP_PROCESS);
+	TASK_GET_PROTOCOL_FLAG (flag_str, "milter",
+			RSPAMD_TASK_PROTOCOL_FLAG_MILTER);
+	TASK_GET_FLAG (flag_str, "bad_unicode",
+			RSPAMD_TASK_FLAG_BAD_UNICODE);
+
+	if (!found) {
+		msg_warn_task ("invalid flag name %s", flag_str);
+		return FALSE;
+	}
+
+	return result;
+}
+
+static gboolean
+rspamd_has_symbol_expr (struct rspamd_task *task,
+					  GArray * args,
+					  void *unused)
+{
+	struct expression_argument *sym_arg;
+	const gchar *symbol_str;
+
+	if (args == NULL) {
+		msg_warn_task ("no parameters to function");
+		return FALSE;
+	}
+
+	sym_arg = &g_array_index (args, struct expression_argument, 0);
+
+	if (sym_arg->type != EXPRESSION_ARGUMENT_NORMAL) {
+		msg_warn_task ("invalid parameter to function");
+		return FALSE;
+	}
+
+	symbol_str = (const gchar *)sym_arg->data;
+
+	if (rspamd_task_find_symbol_result (task, symbol_str, NULL)) {
+		return TRUE;
+	}
+
+	return FALSE;
+}
