@@ -15,20 +15,41 @@
  */
 #include "task.h"
 #include "rspamd.h"
-#include "filter.h"
-#include "protocol.h"
+#include "scan_result.h"
+#include "libserver/protocol.h"
+#include "libserver/protocol_internal.h"
 #include "message.h"
 #include "lua/lua_common.h"
 #include "email_addr.h"
-#include "composites.h"
+#include "src/libserver/composites/composites.h"
 #include "stat_api.h"
 #include "unix-std.h"
 #include "utlist.h"
-#include "contrib/zstd/zstd.h"
 #include "libserver/mempool_vars_internal.h"
+#include "libserver/cfg_file_private.h"
 #include "libmime/lang_detection.h"
+#include "libmime/scan_result_private.h"
+
+#ifdef WITH_JEMALLOC
+#include <jemalloc/jemalloc.h>
+#else
+# if defined(__GLIBC__) && defined(_GNU_SOURCE)
+#  include <malloc.h>
+# endif
+#endif
+
 #include <math.h>
 
+#ifdef SYS_ZSTD
+#  include "zstd.h"
+#else
+#  include "contrib/zstd/zstd.h"
+#endif
+
+__KHASH_IMPL (rspamd_req_headers_hash, static inline,
+		rspamd_ftok_t *, struct rspamd_request_header_chain *, 1,
+				rspamd_ftok_icase_hash, rspamd_ftok_icase_equal)
+
 /*
  * Do not print more than this amount of elts
  */
@@ -40,34 +61,33 @@ rspamd_task_quark (void)
 	return g_quark_from_static_string ("task-error");
 }
 
-static void
-rspamd_request_header_dtor (gpointer p)
-{
-	GPtrArray *ar = p;
-	guint i;
-	rspamd_ftok_t *tok;
-
-	if (ar) {
-		for (i = 0; i < ar->len; i ++) {
-			tok = g_ptr_array_index (ar, i);
-			rspamd_fstring_mapped_ftok_free (tok);
-		}
-
-		g_ptr_array_free (ar, TRUE);
-	}
-}
-
 /*
  * Create new task
  */
 struct rspamd_task *
-rspamd_task_new (struct rspamd_worker *worker, struct rspamd_config *cfg,
-		rspamd_mempool_t *pool,
-		struct rspamd_lang_detector *lang_det)
+rspamd_task_new (struct rspamd_worker *worker,
+				 struct rspamd_config *cfg,
+				 rspamd_mempool_t *pool,
+				 struct rspamd_lang_detector *lang_det,
+				 struct ev_loop *event_loop,
+				 gboolean debug_mem)
 {
 	struct rspamd_task *new_task;
+	rspamd_mempool_t *task_pool;
+	guint flags = 0;
 
-	new_task = g_malloc0 (sizeof (struct rspamd_task));
+	if (pool == NULL) {
+		task_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (),
+				"task", debug_mem ? RSPAMD_MEMPOOL_DEBUG : 0);
+		flags |= RSPAMD_TASK_FLAG_OWN_POOL;
+	}
+	else {
+		task_pool = pool;
+	}
+
+	new_task = rspamd_mempool_alloc0 (task_pool, sizeof (struct rspamd_task));
+	new_task->task_pool = task_pool;
+	new_task->flags = flags;
 	new_task->worker = worker;
 	new_task->lang_det = lang_det;
 
@@ -89,65 +109,17 @@ rspamd_task_new (struct rspamd_worker *worker, struct rspamd_config *cfg,
 		}
 	}
 
-	gettimeofday (&new_task->tv, NULL);
-	new_task->time_real = rspamd_get_ticks (FALSE);
-	new_task->time_virtual = rspamd_get_virtual_ticks ();
+	new_task->event_loop = event_loop;
+	new_task->task_timestamp = ev_time ();
 	new_task->time_real_finish = NAN;
-	new_task->time_virtual_finish = NAN;
-
-	if (pool == NULL) {
-		new_task->task_pool =
-				rspamd_mempool_new (rspamd_mempool_suggest_size (), "task");
-		new_task->flags |= RSPAMD_TASK_FLAG_OWN_POOL;
-	}
-	else {
-		new_task->task_pool = pool;
-	}
-
-	new_task->raw_headers = g_hash_table_new_full (rspamd_strcase_hash,
-			rspamd_strcase_equal, NULL, rspamd_ptr_array_free_hard);
-	new_task->headers_order = g_queue_new ();
-	new_task->request_headers = g_hash_table_new_full (rspamd_ftok_icase_hash,
-			rspamd_ftok_icase_equal, rspamd_fstring_mapped_ftok_free,
-			rspamd_request_header_dtor);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-		(rspamd_mempool_destruct_t) g_hash_table_unref,
-		new_task->request_headers);
-	new_task->reply_headers = g_hash_table_new_full (rspamd_ftok_icase_hash,
-			rspamd_ftok_icase_equal, rspamd_fstring_mapped_ftok_free,
-			rspamd_fstring_mapped_ftok_free);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			(rspamd_mempool_destruct_t) g_hash_table_unref,
-			new_task->reply_headers);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			(rspamd_mempool_destruct_t) g_hash_table_unref,
-			new_task->raw_headers);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			(rspamd_mempool_destruct_t) g_queue_free,
-			new_task->headers_order);
-	new_task->emails = g_hash_table_new (rspamd_email_hash, rspamd_emails_cmp);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			(rspamd_mempool_destruct_t) g_hash_table_unref,
-			new_task->emails);
-	new_task->urls = g_hash_table_new (rspamd_url_hash, rspamd_urls_cmp);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			(rspamd_mempool_destruct_t) g_hash_table_unref,
-			new_task->urls);
-	new_task->parts = g_ptr_array_sized_new (4);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			rspamd_ptr_array_free_hard, new_task->parts);
-	new_task->text_parts = g_ptr_array_sized_new (2);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			rspamd_ptr_array_free_hard, new_task->text_parts);
-	new_task->received = g_ptr_array_sized_new (8);
-	rspamd_mempool_add_destructor (new_task->task_pool,
-			rspamd_ptr_array_free_hard, new_task->received);
 
+	new_task->request_headers = kh_init (rspamd_req_headers_hash);
 	new_task->sock = -1;
-	new_task->flags |= (RSPAMD_TASK_FLAG_MIME|RSPAMD_TASK_FLAG_JSON);
-	new_task->result = rspamd_create_metric_result (new_task);
+	new_task->flags |= (RSPAMD_TASK_FLAG_MIME);
+	/* Default results chain */
+	rspamd_create_metric_result (new_task, NULL, -1);
 
-	new_task->message_id = new_task->queue_id = "undef";
+	new_task->queue_id = "undef";
 	new_task->messages = ucl_object_typed_new (UCL_OBJECT);
 	new_task->lua_cache = g_hash_table_new (rspamd_str_hash, rspamd_str_equal);
 
@@ -158,11 +130,15 @@ rspamd_task_new (struct rspamd_worker *worker, struct rspamd_config *cfg,
 static void
 rspamd_task_reply (struct rspamd_task *task)
 {
+	const ev_tstamp write_timeout = 5.0;
+
 	if (task->fin_callback) {
 		task->fin_callback (task, task->fin_arg);
 	}
 	else {
-		rspamd_protocol_write_reply (task);
+		if (!(task->processed_stages & RSPAMD_TASK_STAGE_REPLIED)) {
+			rspamd_protocol_write_reply (task, write_timeout);
+		}
 	}
 }
 
@@ -210,10 +186,9 @@ rspamd_task_restore (void *arg)
 void
 rspamd_task_free (struct rspamd_task *task)
 {
-	struct rspamd_mime_part *p;
-	struct rspamd_mime_text_part *tp;
 	struct rspamd_email_address *addr;
 	struct rspamd_lua_cached_entry *entry;
+	static guint free_iters = 0;
 	GHashTableIter it;
 	gpointer k, v;
 	guint i;
@@ -221,41 +196,6 @@ rspamd_task_free (struct rspamd_task *task)
 	if (task) {
 		debug_task ("free pointer %p", task);
 
-		for (i = 0; i < task->parts->len; i ++) {
-			p = g_ptr_array_index (task->parts, i);
-
-			if (p->raw_headers) {
-				g_hash_table_unref (p->raw_headers);
-			}
-
-			if (p->headers_order) {
-				g_queue_free (p->headers_order);
-			}
-
-			if (IS_CT_MULTIPART (p->ct)) {
-				if (p->specific.mp.children) {
-					g_ptr_array_free (p->specific.mp.children, TRUE);
-				}
-			}
-		}
-
-		for (i = 0; i < task->text_parts->len; i ++) {
-			tp = g_ptr_array_index (task->text_parts, i);
-
-			if (tp->utf_words) {
-				g_array_free (tp->utf_words, TRUE);
-			}
-			if (tp->normalized_hashes) {
-				g_array_free (tp->normalized_hashes, TRUE);
-			}
-			if (tp->languages) {
-				g_ptr_array_unref (tp->languages);
-			}
-			if (tp->unicode_raw_content) {
-				g_array_free (tp->unicode_raw_content, TRUE);
-			}
-		}
-
 		if (task->rcpt_envelope) {
 			for (i = 0; i < task->rcpt_envelope->len; i ++) {
 				addr = g_ptr_array_index (task->rcpt_envelope, i);
@@ -269,6 +209,14 @@ rspamd_task_free (struct rspamd_task *task)
 			rspamd_email_address_free (task->from_envelope);
 		}
 
+		if (task->from_envelope_orig) {
+			rspamd_email_address_free (task->from_envelope_orig);
+		}
+
+		if (task->meta_words) {
+			g_array_free (task->meta_words, TRUE);
+		}
+
 		ucl_object_unref (task->messages);
 
 		if (task->re_rt) {
@@ -284,6 +232,10 @@ rspamd_task_free (struct rspamd_task *task)
 			ucl_object_unref (task->settings);
 		}
 
+		if (task->settings_elt != NULL) {
+			REF_RELEASE (task->settings_elt);
+		}
+
 		if (task->client_addr) {
 			rspamd_inet_address_free (task->client_addr);
 		}
@@ -296,13 +248,8 @@ rspamd_task_free (struct rspamd_task *task)
 			g_error_free (task->err);
 		}
 
-		if (rspamd_event_pending (&task->timeout_ev, EV_TIMEOUT)) {
-			event_del (&task->timeout_ev);
-		}
-
-		if (task->guard_ev) {
-			event_del (task->guard_ev);
-		}
+		ev_timer_stop (task->event_loop, &task->timeout_ev);
+		ev_io_stop (task->event_loop, &task->guard_ev);
 
 		if (task->sock != -1) {
 			close (task->sock);
@@ -321,20 +268,56 @@ rspamd_task_free (struct rspamd_task *task)
 				g_hash_table_unref (task->lua_cache);
 			}
 
+			if (task->cfg->full_gc_iters && (++free_iters > task->cfg->full_gc_iters)) {
+				/* Perform more expensive cleanup cycle */
+				gsize allocated = 0, active = 0, metadata = 0,
+						resident = 0, mapped = 0, old_lua_mem = 0;
+				gdouble t1, t2;
+
+				old_lua_mem = lua_gc (task->cfg->lua_state, LUA_GCCOUNT, 0);
+				t1 = rspamd_get_ticks (FALSE);
+
+#ifdef WITH_JEMALLOC
+				gsize sz = sizeof (gsize);
+				mallctl ("stats.allocated", &allocated, &sz, NULL, 0);
+				mallctl ("stats.active", &active, &sz, NULL, 0);
+				mallctl ("stats.metadata", &metadata, &sz, NULL, 0);
+				mallctl ("stats.resident", &resident, &sz, NULL, 0);
+				mallctl ("stats.mapped", &mapped, &sz, NULL, 0);
+#else
+# if defined(__GLIBC__) && defined(_GNU_SOURCE)
+				malloc_trim (0);
+# endif
+#endif
+				lua_gc (task->cfg->lua_state, LUA_GCCOLLECT, 0);
+				t2 = rspamd_get_ticks (FALSE);
+
+				msg_notice_task ("perform full gc cycle; memory stats: "
+								 "%Hz allocated, %Hz active, %Hz metadata, %Hz resident, %Hz mapped;"
+								 " lua memory: %z kb -> %d kb; %f ms for gc iter",
+						allocated, active, metadata, resident, mapped,
+						old_lua_mem, lua_gc (task->cfg->lua_state, LUA_GCCOUNT, 0),
+						(t2 - t1) * 1000.0);
+				free_iters = rspamd_time_jitter (0,
+						(gdouble)task->cfg->full_gc_iters / 2);
+			}
+
 			REF_RELEASE (task->cfg);
 		}
 
+		kh_destroy (rspamd_req_headers_hash, task->request_headers);
+		rspamd_message_unref (task->message);
+
 		if (task->flags & RSPAMD_TASK_FLAG_OWN_POOL) {
 			rspamd_mempool_delete (task->task_pool);
 		}
-
-		g_free (task);
 	}
 }
 
 struct rspamd_task_map {
 	gpointer begin;
 	gulong len;
+	gint fd;
 };
 
 static void
@@ -343,11 +326,12 @@ rspamd_task_unmapper (gpointer ud)
 	struct rspamd_task_map *m = ud;
 
 	munmap (m->begin, m->len);
+	close (m->fd);
 }
 
 gboolean
 rspamd_task_load_message (struct rspamd_task *task,
-	struct rspamd_http_message *msg, const gchar *start, gsize len)
+						  struct rspamd_http_message *msg, const gchar *start, gsize len)
 {
 	guint control_len, r;
 	struct ucl_parser *parser;
@@ -417,8 +401,6 @@ rspamd_task_load_message (struct rspamd_task *task,
 			return FALSE;
 		}
 
-		close (fd);
-
 		tok = rspamd_task_get_request_header (task, "shm-offset");
 
 		if (tok) {
@@ -426,8 +408,9 @@ rspamd_task_load_message (struct rspamd_task *task,
 
 			if (offset > (gulong)st.st_size) {
 				msg_err_task ("invalid offset %ul (%ul available) for shm "
-						"segment %s", offset, st.st_size, fp);
+							  "segment %s", offset, (gulong)st.st_size, fp);
 				munmap (map, st.st_size);
+				close (fd);
 
 				return FALSE;
 			}
@@ -442,8 +425,9 @@ rspamd_task_load_message (struct rspamd_task *task,
 
 			if (shmem_size > (gulong)st.st_size) {
 				msg_err_task ("invalid length %ul (%ul available) for %s "
-						"segment %s", shmem_size, st.st_size, ft, fp);
+							  "segment %s", shmem_size, (gulong)st.st_size, ft, fp);
 				munmap (map, st.st_size);
+				close (fd);
 
 				return FALSE;
 			}
@@ -454,9 +438,10 @@ rspamd_task_load_message (struct rspamd_task *task,
 		m = rspamd_mempool_alloc (task->task_pool, sizeof (*m));
 		m->begin = map;
 		m->len = st.st_size;
+		m->fd = fd;
 
-		msg_info_task ("loaded message from shared memory %s (%ul size, %ul offset)",
-				fp, shmem_size, offset);
+		msg_info_task ("loaded message from shared memory %s (%ul size, %ul offset), fd=%d",
+				fp, shmem_size, offset, fd);
 
 		rspamd_mempool_add_destructor (task->task_pool, rspamd_task_unmapper, m);
 
@@ -520,12 +505,12 @@ rspamd_task_load_message (struct rspamd_task *task,
 				return FALSE;
 			}
 
-			close (fd);
 			task->msg.begin = map;
 			task->msg.len = st.st_size;
 			m = rspamd_mempool_alloc (task->task_pool, sizeof (*m));
 			m->begin = map;
 			m->len = st.st_size;
+			m->fd = fd;
 
 			rspamd_mempool_add_destructor (task->task_pool, rspamd_task_unmapper, m);
 		}
@@ -621,7 +606,7 @@ rspamd_task_load_message (struct rspamd_task *task,
 
 				if (zout.pos == zout.size) {
 					/* We need to extend output buffer */
-					zout.size = zout.size * 1.5 + 1.0;
+					zout.size = zout.size * 2 + 1;
 					zout.dst = g_realloc (zout.dst, zout.size);
 				}
 			}
@@ -629,10 +614,10 @@ rspamd_task_load_message (struct rspamd_task *task,
 			rspamd_mempool_add_destructor (task->task_pool, g_free, zout.dst);
 			task->msg.begin = zout.dst;
 			task->msg.len = zout.pos;
-			task->flags |= RSPAMD_TASK_FLAG_COMPRESSED;
+			task->protocol_flags |= RSPAMD_TASK_PROTOCOL_FLAG_COMPRESSED;
 
 			msg_info_task ("loaded message from zstd compressed stream; "
-					"compressed: %ul; uncompressed: %ul",
+						   "compressed: %ul; uncompressed: %ul",
 					(gulong)zin.size, (gulong)zout.pos);
 
 		}
@@ -651,16 +636,20 @@ rspamd_task_load_message (struct rspamd_task *task,
 		task->flags |= RSPAMD_TASK_FLAG_EMPTY;
 	}
 
-	if (task->flags & RSPAMD_TASK_FLAG_HAS_CONTROL) {
-		/* We have control chunk, so we need to process it separately */
-		if (task->msg.len < task->message_len) {
+	if (task->protocol_flags & RSPAMD_TASK_PROTOCOL_FLAG_HAS_CONTROL) {
+		rspamd_ftok_t *hv = rspamd_task_get_request_header (task, MLEN_HEADER);
+		gulong message_len = 0;
+
+		if (!hv || !rspamd_strtoul (hv->begin, hv->len, &message_len) ||
+				task->msg.len < message_len) {
 			msg_warn_task ("message has invalid message length: %ul and total len: %ul",
-					task->message_len, task->msg.len);
+					message_len, task->msg.len);
 			g_set_error (&task->err, rspamd_task_quark(), RSPAMD_PROTOCOL_ERROR,
 					"Invalid length");
 			return FALSE;
 		}
-		control_len = task->msg.len - task->message_len;
+
+		control_len = task->msg.len - message_len;
 
 		if (control_len > 0) {
 			parser = ucl_parser_new (UCL_PARSER_KEY_LOWERCASE);
@@ -720,7 +709,7 @@ gboolean
 rspamd_task_process (struct rspamd_task *task, guint stages)
 {
 	gint st;
-	gboolean ret = TRUE;
+	gboolean ret = TRUE, all_done = TRUE;
 	GError *stat_error = NULL;
 
 	/* Avoid nested calls */
@@ -737,26 +726,25 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 	st = rspamd_task_select_processing_stage (task, stages);
 
 	switch (st) {
+	case RSPAMD_TASK_STAGE_CONNFILTERS:
+		all_done = rspamd_symcache_process_symbols (task, task->cfg->cache, st);
+		break;
+
 	case RSPAMD_TASK_STAGE_READ_MESSAGE:
 		if (!rspamd_message_parse (task)) {
 			ret = FALSE;
 		}
 		break;
 
-	case RSPAMD_TASK_STAGE_PRE_FILTERS:
-		rspamd_symbols_cache_process_symbols (task, task->cfg->cache,
-				RSPAMD_TASK_STAGE_PRE_FILTERS);
-		break;
-
 	case RSPAMD_TASK_STAGE_PROCESS_MESSAGE:
 		if (!(task->flags & RSPAMD_TASK_FLAG_SKIP_PROCESS)) {
 			rspamd_message_process (task);
 		}
 		break;
 
+	case RSPAMD_TASK_STAGE_PRE_FILTERS:
 	case RSPAMD_TASK_STAGE_FILTERS:
-		rspamd_symbols_cache_process_symbols (task, task->cfg->cache,
-				RSPAMD_TASK_STAGE_FILTERS);
+		all_done = rspamd_symcache_process_symbols (task, task->cfg->cache, st);
 		break;
 
 	case RSPAMD_TASK_STAGE_CLASSIFIERS:
@@ -772,14 +760,15 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 		break;
 
 	case RSPAMD_TASK_STAGE_COMPOSITES:
-		rspamd_make_composites (task);
+		rspamd_composites_process_task (task);
+		task->result->nresults_postfilters = task->result->nresults;
 		break;
 
 	case RSPAMD_TASK_STAGE_POST_FILTERS:
-		rspamd_symbols_cache_process_symbols (task, task->cfg->cache,
-				RSPAMD_TASK_STAGE_POST_FILTERS);
+		all_done = rspamd_symcache_process_symbols (task, task->cfg->cache,
+				st);
 
-		if ((task->flags & RSPAMD_TASK_FLAG_LEARN_AUTO) &&
+		if (all_done && (task->flags & RSPAMD_TASK_FLAG_LEARN_AUTO) &&
 				!RSPAMD_TASK_IS_EMPTY (task) &&
 				!(task->flags & (RSPAMD_TASK_FLAG_LEARN_SPAM|RSPAMD_TASK_FLAG_LEARN_HAM))) {
 			rspamd_stat_check_autolearn (task);
@@ -799,10 +788,17 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 					if (stat_error == NULL) {
 						g_set_error (&stat_error,
 								g_quark_from_static_string ("stat"), 500,
-								"Unknown statistics error");
+								"Unknown statistics error, found on stage %s;"
+								" classifier: %s",
+								rspamd_task_stage_name (st), task->classifier);
 					}
 
-					msg_err_task ("learn error: %e", stat_error);
+					if (stat_error->code >= 400) {
+						msg_err_task ("learn error: %e", stat_error);
+					}
+					else {
+						msg_notice_task ("skip learning: %e", stat_error);
+					}
 
 					if (!(task->flags & RSPAMD_TASK_FLAG_LEARN_AUTO)) {
 						task->err = stat_error;
@@ -823,12 +819,22 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 		}
 		break;
 	case RSPAMD_TASK_STAGE_COMPOSITES_POST:
-		/* Second run of composites processing before idempotent filters */
-		rspamd_make_composites (task);
+		/* Second run of composites processing before idempotent filters (if needed) */
+		if (task->result->nresults_postfilters != task->result->nresults) {
+			rspamd_composites_process_task (task);
+		}
+		else {
+			msg_debug_task ("skip second run of composites as the result has not been changed");
+		}
 		break;
+
 	case RSPAMD_TASK_STAGE_IDEMPOTENT:
-		rspamd_symbols_cache_process_symbols (task, task->cfg->cache,
-				RSPAMD_TASK_STAGE_IDEMPOTENT);
+		/* Stop task timeout */
+		if (ev_can_stop (&task->timeout_ev)) {
+			ev_timer_stop (task->event_loop, &task->timeout_ev);
+		}
+
+		all_done = rspamd_symcache_process_symbols (task, task->cfg->cache, st);
 		break;
 
 	case RSPAMD_TASK_STAGE_DONE:
@@ -858,17 +864,24 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 		return ret;
 	}
 
-	if (rspamd_session_events_pending (task->s) != 0) {
-		/* We have events pending, so we consider this stage as incomplete */
-		msg_debug_task ("need more work on stage %d", st);
-	}
-	else {
-		/* Mark the current stage as done and go to the next stage */
-		msg_debug_task ("completed stage %d", st);
-		task->processed_stages |= st;
+	if (ret) {
+		if (rspamd_session_events_pending (task->s) != 0) {
+			/* We have events pending, so we consider this stage as incomplete */
+			msg_debug_task ("need more work on stage %d", st);
+		}
+		else {
+			if (all_done) {
+				/* Mark the current stage as done and go to the next stage */
+				msg_debug_task ("completed stage %d", st);
+				task->processed_stages |= st;
+			}
+			else {
+				msg_debug_task ("need more processing on stage %d", st);
+			}
 
-		/* Tail recursion */
-		return rspamd_task_process (task, stages);
+			/* Tail recursion */
+			return rspamd_task_process (task, stages);
+		}
 	}
 
 	return ret;
@@ -905,6 +918,7 @@ rspamd_task_get_principal_recipient (struct rspamd_task *task)
 {
 	const gchar *val;
 	struct rspamd_email_address *addr;
+	guint i;
 
 	val = rspamd_mempool_get_variable (task->task_pool,
 			RSPAMD_MEMPOOL_PRINCIPAL_RECIPIENT);
@@ -918,20 +932,22 @@ rspamd_task_get_principal_recipient (struct rspamd_task *task)
 				strlen (task->deliver_to));
 	}
 	if (task->rcpt_envelope != NULL) {
-		addr = g_ptr_array_index (task->rcpt_envelope, 0);
 
-		if (addr->addr) {
-			return rspamd_task_cache_principal_recipient (task, addr->addr,
-					addr->addr_len);
+		PTR_ARRAY_FOREACH (task->rcpt_envelope, i, addr) {
+			if (addr->addr && !(addr->flags & RSPAMD_EMAIL_ADDR_ORIGINAL)) {
+				return rspamd_task_cache_principal_recipient (task, addr->addr,
+						addr->addr_len);
+			}
 		}
 	}
 
-	if (task->rcpt_mime != NULL && task->rcpt_mime->len > 0) {
-		addr = g_ptr_array_index (task->rcpt_mime, 0);
-
-		if (addr->addr) {
-			return rspamd_task_cache_principal_recipient (task, addr->addr,
-					addr->addr_len);
+	GPtrArray *rcpt_mime = MESSAGE_FIELD_CHECK (task, rcpt_mime);
+	if (rcpt_mime != NULL && rcpt_mime->len > 0) {
+		PTR_ARRAY_FOREACH (rcpt_mime, i, addr) {
+			if (addr->addr && !(addr->flags & RSPAMD_EMAIL_ADDR_ORIGINAL)) {
+				return rspamd_task_cache_principal_recipient (task, addr->addr,
+						addr->addr_len);
+			}
 		}
 	}
 
@@ -964,7 +980,8 @@ rspamd_task_log_check_condition (struct rspamd_task *task,
 
 	switch (lf->type) {
 	case RSPAMD_LOG_MID:
-		if (task->message_id && strcmp (task->message_id, "undef") != 0) {
+		if (MESSAGE_FIELD_CHECK (task, message_id) &&
+			strcmp (MESSAGE_FIELD (task, message_id) , "undef") != 0) {
 			ret = TRUE;
 		}
 		break;
@@ -991,7 +1008,8 @@ rspamd_task_log_check_condition (struct rspamd_task *task,
 		break;
 	case RSPAMD_LOG_MIME_RCPT:
 	case RSPAMD_LOG_MIME_RCPTS:
-		if (task->rcpt_mime && task->rcpt_mime->len > 0) {
+		if (MESSAGE_FIELD_CHECK (task, rcpt_mime) &&
+			MESSAGE_FIELD (task, rcpt_mime)->len > 0) {
 			ret = TRUE;
 		}
 		break;
@@ -1001,7 +1019,8 @@ rspamd_task_log_check_condition (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_MIME_FROM:
-		if (task->from_mime && task->from_mime->len > 0) {
+		if (MESSAGE_FIELD_CHECK (task, from_mime) &&
+			MESSAGE_FIELD (task, from_mime)->len > 0) {
 			ret = TRUE;
 		}
 		break;
@@ -1015,6 +1034,11 @@ rspamd_task_log_check_condition (struct rspamd_task *task,
 			ret = TRUE;
 		}
 		break;
+	case RSPAMD_LOG_SETTINGS_ID:
+		if (task->settings_elt) {
+			ret = TRUE;
+		}
+		break;
 	default:
 		ret = TRUE;
 		break;
@@ -1044,22 +1068,34 @@ rspamd_task_compare_log_sym (gconstpointer a, gconstpointer b)
 	return (w2 - w1) * 1000.0;
 }
 
+static gint
+rspamd_task_compare_log_group (gconstpointer a, gconstpointer b)
+{
+	const struct rspamd_symbols_group *s1 = *(const struct rspamd_symbols_group **)a,
+			*s2 = *(const struct rspamd_symbols_group **)b;
+
+	return strcmp (s1->name, s2->name);
+}
+
+
 static rspamd_ftok_t
 rspamd_task_log_metric_res (struct rspamd_task *task,
 		struct rspamd_log_format *lf)
 {
 	static gchar scorebuf[32];
 	rspamd_ftok_t res = {.begin = NULL, .len = 0};
-	struct rspamd_metric_result *mres;
+	struct rspamd_scan_result *mres;
 	gboolean first = TRUE;
 	rspamd_fstring_t *symbuf;
 	struct rspamd_symbol_result *sym;
 	GPtrArray *sorted_symbols;
-	enum rspamd_action_type act;
+	struct rspamd_action *act;
+	struct rspamd_symbols_group *gr;
 	guint i, j;
+	khiter_t k;
 
 	mres = task->result;
-	act = rspamd_check_action_metric (task, mres);
+	act = rspamd_check_action_metric (task, NULL, NULL);
 
 	if (mres != NULL) {
 		switch (lf->type) {
@@ -1067,7 +1103,7 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 			if (RSPAMD_TASK_IS_SKIPPED (task)) {
 				res.begin = "S";
 			}
-			else if (act == METRIC_ACTION_REJECT) {
+			else if (!(act->flags & RSPAMD_ACTION_HAM)) {
 				res.begin = "T";
 			}
 			else {
@@ -1077,7 +1113,7 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 			res.len = 1;
 			break;
 		case RSPAMD_LOG_ACTION:
-			res.begin = rspamd_action_to_str (act);
+			res.begin = act->name;
 			res.len = strlen (res.begin);
 			break;
 		case RSPAMD_LOG_SCORES:
@@ -1089,7 +1125,7 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 			symbuf = rspamd_fstring_sized_new (128);
 			sorted_symbols = g_ptr_array_sized_new (kh_size (mres->symbols));
 
-			kh_foreach_value_ptr (mres->symbols, sym, {
+			kh_foreach_value (mres->symbols, sym, {
 				if (!(sym->flags & RSPAMD_SYMBOL_RESULT_IGNORED)) {
 					g_ptr_array_add (sorted_symbols, (gpointer)sym);
 				}
@@ -1120,7 +1156,8 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 						j = 0;
 
 						DL_FOREACH (sym->opts_head, opt) {
-							rspamd_printf_fstring (&symbuf, "%s;", opt->option);
+							rspamd_printf_fstring (&symbuf, "%*s;",
+									(gint)opt->optlen, opt->option);
 
 							if (j >= max_log_elts) {
 								rspamd_printf_fstring (&symbuf, "...;");
@@ -1142,6 +1179,53 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 			rspamd_mempool_add_destructor (task->task_pool,
 					(rspamd_mempool_destruct_t)rspamd_fstring_free,
 					symbuf);
+			rspamd_mempool_notify_alloc (task->task_pool, symbuf->len);
+			res.begin = symbuf->str;
+			res.len = symbuf->len;
+			break;
+
+		case RSPAMD_LOG_GROUPS:
+		case RSPAMD_LOG_PUBLIC_GROUPS:
+
+			symbuf = rspamd_fstring_sized_new (128);
+			sorted_symbols = g_ptr_array_sized_new (kh_size (mres->sym_groups));
+
+			kh_foreach_key (mres->sym_groups, gr,{
+				if (!(gr->flags & RSPAMD_SYMBOL_GROUP_PUBLIC)) {
+					if (lf->type == RSPAMD_LOG_PUBLIC_GROUPS) {
+						continue;
+					}
+				}
+
+				g_ptr_array_add (sorted_symbols, gr);
+			});
+
+			g_ptr_array_sort (sorted_symbols, rspamd_task_compare_log_group);
+
+			for (i = 0; i < sorted_symbols->len; i++) {
+				gr = g_ptr_array_index (sorted_symbols, i);
+
+				if (first) {
+					rspamd_printf_fstring (&symbuf, "%s", gr->name);
+				}
+				else {
+					rspamd_printf_fstring (&symbuf, ",%s", gr->name);
+				}
+
+				k = kh_get (rspamd_symbols_group_hash, mres->sym_groups, gr);
+
+				rspamd_printf_fstring (&symbuf, "(%.2f)",
+						kh_value (mres->sym_groups, k));
+
+				first = FALSE;
+			}
+
+			g_ptr_array_free (sorted_symbols, TRUE);
+
+			rspamd_mempool_add_destructor (task->task_pool,
+					(rspamd_mempool_destruct_t) rspamd_fstring_free,
+					symbuf);
+			rspamd_mempool_notify_alloc (task->task_pool, symbuf->len);
 			res.begin = symbuf->str;
 			res.len = symbuf->len;
 			break;
@@ -1202,22 +1286,39 @@ rspamd_task_write_ialist (struct rspamd_task *task,
 	rspamd_fstring_t *res = logbuf, *varbuf;
 	rspamd_ftok_t var = {.begin = NULL, .len = 0};
 	struct rspamd_email_address *addr;
-	gint i, nchars = 0, cur_chars;
+	gint i, nchars = 0, wr = 0, cur_chars;
+	gboolean has_orig = FALSE;
 
-	if (lim <= 0) {
+	if (addrs && lim <= 0) {
 		lim = addrs->len;
 	}
 
+	PTR_ARRAY_FOREACH (addrs, i, addr) {
+		if (addr->flags & RSPAMD_EMAIL_ADDR_ORIGINAL) {
+			has_orig = TRUE;
+			break;
+		}
+	}
+
 	varbuf = rspamd_fstring_new ();
 
 	PTR_ARRAY_FOREACH (addrs, i, addr) {
-		if (i >= lim) {
+		if (wr >= lim) {
 			break;
 		}
+
+		if (has_orig) {
+			/* Report merely original addresses */
+			if (!(addr->flags & RSPAMD_EMAIL_ADDR_ORIGINAL)) {
+				continue;
+			}
+		}
+
 		cur_chars = addr->addr_len;
 		varbuf = rspamd_fstring_append (varbuf, addr->addr,
 				cur_chars);
 		nchars += cur_chars;
+		wr ++;
 
 		if (varbuf->len > 0) {
 			if (i != lim - 1) {
@@ -1225,7 +1326,7 @@ rspamd_task_write_ialist (struct rspamd_task *task,
 			}
 		}
 
-		if (i >= max_log_elts || nchars >= max_log_elts * 10) {
+		if (wr >= max_log_elts || nchars >= max_log_elts * 10) {
 			varbuf = rspamd_fstring_append (varbuf, "...", 3);
 			break;
 		}
@@ -1303,8 +1404,8 @@ rspamd_task_log_variable (struct rspamd_task *task,
 	switch (lf->type) {
 	/* String vars */
 	case RSPAMD_LOG_MID:
-		if (task->message_id) {
-			var.begin = task->message_id;
+		if (MESSAGE_FIELD_CHECK (task, message_id)) {
+			var.begin = MESSAGE_FIELD (task, message_id);
 			var.len = strlen (var.begin);
 		}
 		else {
@@ -1354,14 +1455,14 @@ rspamd_task_log_variable (struct rspamd_task *task,
 		var.begin = numbuf;
 		break;
 	case RSPAMD_LOG_TIME_REAL:
-		var.begin = rspamd_log_check_time (task->time_real,
+		var.begin = rspamd_log_check_time (task->task_timestamp,
 				task->time_real_finish,
 				task->cfg->clock_res);
 		var.len = strlen (var.begin);
 		break;
 	case RSPAMD_LOG_TIME_VIRTUAL:
-		var.begin = rspamd_log_check_time (task->time_virtual,
-				task->time_virtual_finish,
+		var.begin = rspamd_log_check_time (task->task_timestamp,
+				task->time_real_finish,
 				task->cfg->clock_res);
 		var.len = strlen (var.begin);
 		break;
@@ -1373,8 +1474,11 @@ rspamd_task_log_variable (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_MIME_FROM:
-		if (task->from_mime) {
-			return rspamd_task_write_ialist (task, task->from_mime, 1, lf,
+		if (MESSAGE_FIELD_CHECK (task, from_mime)) {
+			return rspamd_task_write_ialist (task,
+					MESSAGE_FIELD (task, from_mime),
+					1,
+					lf,
 					logbuf);
 		}
 		break;
@@ -1385,8 +1489,11 @@ rspamd_task_log_variable (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_MIME_RCPT:
-		if (task->rcpt_mime) {
-			return rspamd_task_write_ialist (task, task->rcpt_mime, 1, lf,
+		if (MESSAGE_FIELD_CHECK (task, rcpt_mime)) {
+			return rspamd_task_write_ialist (task,
+					MESSAGE_FIELD (task, rcpt_mime),
+					1,
+					lf,
 					logbuf);
 		}
 		break;
@@ -1397,15 +1504,25 @@ rspamd_task_log_variable (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_MIME_RCPTS:
-		if (task->rcpt_mime) {
-			return rspamd_task_write_ialist (task, task->rcpt_mime, -1, lf,
+		if (MESSAGE_FIELD_CHECK (task, rcpt_mime)) {
+			return rspamd_task_write_ialist (task,
+					MESSAGE_FIELD (task, rcpt_mime),
+					-1, /* All addresses */
+					lf,
 					logbuf);
 		}
 		break;
 	case RSPAMD_LOG_DIGEST:
-		var.len = rspamd_snprintf (numbuf, sizeof (numbuf), "%*xs",
-				(gint)sizeof (task->digest), task->digest);
-		var.begin = numbuf;
+		if (task->message) {
+			var.len = rspamd_snprintf (numbuf, sizeof (numbuf), "%*xs",
+					(gint) sizeof (MESSAGE_FIELD (task, digest)),
+					MESSAGE_FIELD (task, digest));
+			var.begin = numbuf;
+		}
+		else {
+			var.begin = undef;
+			var.len = sizeof (undef) - 1;
+		}
 		break;
 	case RSPAMD_LOG_FILENAME:
 		if (task->msg.fpath) {
@@ -1424,14 +1541,17 @@ rspamd_task_log_variable (struct rspamd_task *task,
 			if (!isnan (pr->target_score)) {
 				var.len = rspamd_snprintf (numbuf, sizeof (numbuf),
 						"%s \"%s\"; score=%.2f (set by %s)",
-						rspamd_action_to_str (pr->action),
-						pr->message, pr->target_score, pr->module);
+						pr->action->name,
+						pr->message,
+						pr->target_score,
+						pr->module);
 			}
 			else {
 				var.len = rspamd_snprintf (numbuf, sizeof (numbuf),
 						"%s \"%s\"; score=nan (set by %s)",
-						rspamd_action_to_str (pr->action),
-						pr->message, pr->module);
+						pr->action->name,
+						pr->message,
+						pr->module);
 			}
 			var.begin = numbuf;
 		}
@@ -1440,6 +1560,28 @@ rspamd_task_log_variable (struct rspamd_task *task,
 			var.len = sizeof (undef) - 1;
 		}
 		break;
+	case RSPAMD_LOG_SETTINGS_ID:
+		if (task->settings_elt) {
+			var.begin = task->settings_elt->name;
+			var.len = strlen (task->settings_elt->name);
+		}
+		else {
+			var.begin = undef;
+			var.len = sizeof (undef) - 1;
+		}
+		break;
+	case RSPAMD_LOG_MEMPOOL_SIZE:
+		var.len = rspamd_snprintf (numbuf, sizeof (numbuf),
+				"%Hz",
+				rspamd_mempool_get_used_size (task->task_pool));
+		var.begin = numbuf;
+		break;
+	case RSPAMD_LOG_MEMPOOL_WASTE:
+		var.len = rspamd_snprintf (numbuf, sizeof (numbuf),
+				"%Hz",
+				rspamd_mempool_get_wasted_size (task->task_pool));
+		var.begin = numbuf;
+		break;
 	default:
 		var = rspamd_task_log_metric_res (task, lf);
 		break;
@@ -1467,6 +1609,7 @@ rspamd_task_write_log (struct rspamd_task *task)
 
 	if (task->cfg->log_format == NULL ||
 			(task->flags & RSPAMD_TASK_FLAG_NO_LOG)) {
+		msg_debug_task ("skip logging due to no log flag");
 		return;
 	}
 
@@ -1517,9 +1660,9 @@ rspamd_task_write_log (struct rspamd_task *task)
 }
 
 gdouble
-rspamd_task_get_required_score (struct rspamd_task *task, struct rspamd_metric_result *m)
+rspamd_task_get_required_score (struct rspamd_task *task, struct rspamd_scan_result *m)
 {
-	guint i;
+	gint i;
 
 	if (m == NULL) {
 		m = task->result;
@@ -1529,9 +1672,13 @@ rspamd_task_get_required_score (struct rspamd_task *task, struct rspamd_metric_r
 		}
 	}
 
-	for (i = METRIC_ACTION_REJECT; i < METRIC_ACTION_NOACTION; i ++) {
-		if (!isnan (m->actions_limits[i])) {
-			return m->actions_limits[i];
+	for (i = m->nactions - 1; i >= 0; i --) {
+		struct rspamd_action_result *action_lim = &m->actions_limits[i];
+
+
+		if (!isnan (action_lim->cur_limit) &&
+				!(action_lim->action->flags & (RSPAMD_ACTION_NO_THRESHOLD|RSPAMD_ACTION_HAM))) {
+			return m->actions_limits[i].cur_limit;
 		}
 	}
 
@@ -1542,32 +1689,33 @@ rspamd_ftok_t *
 rspamd_task_get_request_header (struct rspamd_task *task,
 		const gchar *name)
 {
-	GPtrArray *ret;
-	rspamd_ftok_t srch;
-
-	srch.begin = (gchar *)name;
-	srch.len = strlen (name);
-
-	ret = g_hash_table_lookup (task->request_headers, &srch);
+	struct rspamd_request_header_chain *ret =
+			rspamd_task_get_request_header_multiple (task, name);
 
 	if (ret) {
-		return (rspamd_ftok_t *)g_ptr_array_index (ret, 0);
+		return ret->hdr;
 	}
 
 	return NULL;
 }
 
-GPtrArray*
+struct rspamd_request_header_chain *
 rspamd_task_get_request_header_multiple (struct rspamd_task *task,
 		const gchar *name)
 {
-	GPtrArray *ret;
+	struct rspamd_request_header_chain *ret = NULL;
 	rspamd_ftok_t srch;
+	khiter_t k;
 
 	srch.begin = (gchar *)name;
 	srch.len = strlen (name);
 
-	ret = g_hash_table_lookup (task->request_headers, &srch);
+	k = kh_get (rspamd_req_headers_hash, task->request_headers,
+			&srch);
+
+	if (k != kh_end (task->request_headers)) {
+		ret = kh_value (task->request_headers, k);
+	}
 
 	return ret;
 }
@@ -1577,20 +1725,30 @@ void
 rspamd_task_add_request_header (struct rspamd_task *task,
 		rspamd_ftok_t *name, rspamd_ftok_t *value)
 {
-	GPtrArray *ret;
 
-	ret = g_hash_table_lookup (task->request_headers, name);
+	khiter_t k;
+	gint res;
+	struct rspamd_request_header_chain *chain, *nchain;
 
-	if (ret) {
-		g_ptr_array_add (ret, value);
+	k = kh_put (rspamd_req_headers_hash, task->request_headers,
+		name, &res);
 
-		/* We need to free name token */
-		rspamd_fstring_mapped_ftok_free (name);
+	if (res == 0) {
+		/* Existing name */
+		nchain = rspamd_mempool_alloc (task->task_pool, sizeof (*nchain));
+		nchain->hdr = value;
+		nchain->next = NULL;
+		chain = kh_value (task->request_headers, k);
+
+		/* Slow but OK here */
+		LL_APPEND (chain, nchain);
 	}
 	else {
-		ret = g_ptr_array_sized_new (2);
-		g_ptr_array_add (ret, value);
-		g_hash_table_replace (task->request_headers, name, ret);
+		nchain = rspamd_mempool_alloc (task->task_pool, sizeof (*nchain));
+		nchain->hdr = value;
+		nchain->next = NULL;
+
+		kh_value (task->request_headers, k) = nchain;
 	}
 }
 
@@ -1646,11 +1804,186 @@ gboolean
 rspamd_task_set_finish_time (struct rspamd_task *task)
 {
 	if (isnan (task->time_real_finish)) {
-		task->time_real_finish = rspamd_get_ticks (FALSE);
-		task->time_virtual_finish = rspamd_get_virtual_ticks ();
+		task->time_real_finish = ev_time ();
 
 		return TRUE;
 	}
 
 	return FALSE;
-}
\ No newline at end of file
+}
+
+const gchar *
+rspamd_task_stage_name (enum rspamd_task_stage stg)
+{
+	const gchar *ret = "unknown stage";
+
+	switch (stg) {
+	case RSPAMD_TASK_STAGE_CONNECT:
+		ret = "connect";
+		break;
+	case RSPAMD_TASK_STAGE_CONNFILTERS:
+		ret = "connection_filter";
+		break;
+	case RSPAMD_TASK_STAGE_READ_MESSAGE:
+		ret = "read_message";
+		break;
+	case RSPAMD_TASK_STAGE_PRE_FILTERS:
+		ret = "prefilters";
+		break;
+	case RSPAMD_TASK_STAGE_PROCESS_MESSAGE:
+		ret = "process_message";
+		break;
+	case RSPAMD_TASK_STAGE_FILTERS:
+		ret = "filters";
+		break;
+	case RSPAMD_TASK_STAGE_CLASSIFIERS_PRE:
+		ret = "classifiers_pre";
+		break;
+	case RSPAMD_TASK_STAGE_CLASSIFIERS:
+		ret = "classifiers";
+		break;
+	case RSPAMD_TASK_STAGE_CLASSIFIERS_POST:
+		ret = "classifiers_post";
+		break;
+	case RSPAMD_TASK_STAGE_COMPOSITES:
+		ret = "composites";
+		break;
+	case RSPAMD_TASK_STAGE_POST_FILTERS:
+		ret = "postfilters";
+		break;
+	case RSPAMD_TASK_STAGE_LEARN_PRE:
+		ret = "learn_pre";
+		break;
+	case RSPAMD_TASK_STAGE_LEARN:
+		ret = "learn";
+		break;
+	case RSPAMD_TASK_STAGE_LEARN_POST:
+		ret = "learn_post";
+		break;
+	case RSPAMD_TASK_STAGE_COMPOSITES_POST:
+		ret = "composites_post";
+		break;
+	case RSPAMD_TASK_STAGE_IDEMPOTENT:
+		ret = "idempotent";
+		break;
+	case RSPAMD_TASK_STAGE_DONE:
+		ret = "done";
+		break;
+	case RSPAMD_TASK_STAGE_REPLIED:
+		ret = "replied";
+		break;
+	default:
+		break;
+	}
+
+	return ret;
+}
+
+void
+rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
+{
+	struct rspamd_task *task = (struct rspamd_task *)w->data;
+
+	if (!(task->processed_stages & RSPAMD_TASK_STAGE_FILTERS)) {
+		ev_now_update_if_cheap (task->event_loop);
+		msg_info_task ("processing of task time out: %.1fs spent; %.1fs limit; "
+					   "forced processing",
+				ev_now (task->event_loop) - task->task_timestamp,
+				w->repeat);
+
+		if (task->cfg->soft_reject_on_timeout) {
+			struct rspamd_action *action, *soft_reject;
+
+			action = rspamd_check_action_metric (task, NULL, NULL);
+
+			if (action->action_type != METRIC_ACTION_REJECT) {
+				soft_reject = rspamd_config_get_action_by_type (task->cfg,
+						METRIC_ACTION_SOFT_REJECT);
+				rspamd_add_passthrough_result (task,
+						soft_reject,
+						0,
+						NAN,
+						"timeout processing message",
+						"task timeout",
+						0, NULL);
+			}
+		}
+
+		ev_timer_again (EV_A_ w);
+		task->processed_stages |= RSPAMD_TASK_STAGE_FILTERS;
+		rspamd_session_cleanup (task->s);
+		rspamd_task_process (task, RSPAMD_TASK_PROCESS_ALL);
+		rspamd_session_pending (task->s);
+	}
+	else {
+		/* Postprocessing timeout */
+		msg_info_task ("post-processing of task time out: %.1f second spent; forced processing",
+				ev_now (task->event_loop) - task->task_timestamp);
+
+		if (task->cfg->soft_reject_on_timeout) {
+			struct rspamd_action *action, *soft_reject;
+
+			action = rspamd_check_action_metric (task, NULL, NULL);
+
+			if (action->action_type != METRIC_ACTION_REJECT) {
+				soft_reject = rspamd_config_get_action_by_type (task->cfg,
+						METRIC_ACTION_SOFT_REJECT);
+				rspamd_add_passthrough_result (task,
+						soft_reject,
+						0,
+						NAN,
+						"timeout post-processing message",
+						"task timeout",
+						0, NULL);
+			}
+		}
+
+		ev_timer_stop (EV_A_ w);
+		task->processed_stages |= RSPAMD_TASK_STAGE_DONE;
+		rspamd_session_cleanup (task->s);
+		rspamd_task_process (task, RSPAMD_TASK_PROCESS_ALL);
+		rspamd_session_pending (task->s);
+	}
+}
+
+void
+rspamd_worker_guard_handler (EV_P_ ev_io *w, int revents)
+{
+	struct rspamd_task *task = (struct rspamd_task *)w->data;
+	gchar fake_buf[1024];
+	gssize r;
+
+	r = read (w->fd, fake_buf, sizeof (fake_buf));
+
+	if (r > 0) {
+		msg_warn_task ("received extra data after task is loaded, ignoring");
+	}
+	else {
+		if (r == 0) {
+			/*
+			 * Poor man approach, that might break things in case of
+			 * shutdown (SHUT_WR) but sockets are so bad that there's no
+			 * reliable way to distinguish between shutdown(SHUT_WR) and
+			 * close.
+			 */
+			if (task->cmd != CMD_CHECK_V2 && task->cfg->enable_shutdown_workaround) {
+				msg_info_task ("workaround for shutdown enabled, please update "
+							   "your client, this support might be removed in future");
+				shutdown (w->fd, SHUT_RD);
+				ev_io_stop (task->event_loop, &task->guard_ev);
+			}
+			else {
+				msg_err_task ("the peer has closed connection unexpectedly");
+				rspamd_session_destroy (task->s);
+			}
+		}
+		else if (errno != EAGAIN) {
+			msg_err_task ("the peer has closed connection unexpectedly: %s",
+					strerror (errno));
+			rspamd_session_destroy (task->s);
+		}
+		else {
+			return;
+		}
+	}
+}
