@@ -12,12 +12,29 @@
 #include <netdb.h>
 #endif
 
-#include <event.h>
+#include "contrib/libev/ev.h"
 #include <time.h>
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 struct rspamd_config;
-struct rspamd_main;
-struct workq;
+
+enum rspamd_exception_type {
+	RSPAMD_EXCEPTION_NEWLINE = 0,
+	RSPAMD_EXCEPTION_URL,
+	RSPAMD_EXCEPTION_GENERIC,
+};
+/**
+ * Structure to point exception in text from processing
+ */
+struct rspamd_process_exception {
+	goffset pos;
+	guint len;
+	gpointer ptr;
+	enum rspamd_exception_type type;
+};
 
 /**
  * Create generic socket
@@ -28,10 +45,12 @@ struct workq;
  * @return socket FD or -1 in case of error
  */
 gint rspamd_socket_create (gint af, gint type, gint protocol, gboolean async);
+
 /*
  * Create socket and bind or connect it to specified address and port
  */
 gint rspamd_socket_tcp (struct addrinfo *, gboolean is_server, gboolean async);
+
 /*
  * Create socket and bind or connect it to specified address and port
  */
@@ -41,10 +60,10 @@ gint rspamd_socket_udp (struct addrinfo *, gboolean is_server, gboolean async);
  * Create and bind or connect unix socket
  */
 gint rspamd_socket_unix (const gchar *,
-	struct sockaddr_un *,
-	gint type,
-	gboolean is_server,
-	gboolean async);
+						 struct sockaddr_un *,
+						 gint type,
+						 gboolean is_server,
+						 gboolean async);
 
 /**
  * Make a universal socket
@@ -56,37 +75,19 @@ gint rspamd_socket_unix (const gchar *,
  * @param try_resolve try name resolution for a socket (BLOCKING)
  */
 gint rspamd_socket (const gchar *credits, guint16 port, gint type,
-	gboolean async, gboolean is_server, gboolean try_resolve);
+					gboolean async, gboolean is_server, gboolean try_resolve);
+
 
-/**
- * Make a universal sockets
- * @param credits host, ip or path to unix socket (several items may be separated by ',')
- * @param port port (used for network sockets)
- * @param type type of socket (SO_STREAM or SO_DGRAM)
- * @param async make this socket asynced
- * @param is_server make this socket as server socket
- * @param try_resolve try name resolution for a socket (BLOCKING)
- */
-GList * rspamd_sockets_list (const gchar *credits,
-	guint16 port,
-	gint type,
-	gboolean async,
-	gboolean is_server,
-	gboolean try_resolve);
 /*
  * Create socketpair
  */
-gboolean rspamd_socketpair (gint pair[2], gboolean is_stream);
-
-/*
- * Write pid to file
- */
-gint rspamd_write_pid (struct rspamd_main *);
+gboolean rspamd_socketpair (gint pair[2], gint af);
 
 /*
  * Make specified socket non-blocking
  */
 gint rspamd_socket_nonblocking (gint);
+
 /*
  * Make specified socket blocking
  */
@@ -101,24 +102,24 @@ gint rspamd_socket_poll (gint fd, gint timeout, short events);
  * Init signals
  */
 #ifdef HAVE_SA_SIGINFO
-void rspamd_signals_init (struct sigaction *sa, void (*sig_handler)(gint,
-	siginfo_t *,
-	void *));
+
+void rspamd_signals_init (struct sigaction *sa, void (*sig_handler) (gint,
+																	 siginfo_t *,
+																	 void *));
+
 #else
 void rspamd_signals_init (struct sigaction *sa, void (*sig_handler)(gint));
 #endif
 
-/*
- * Send specified signal to each worker
- */
-void rspamd_pass_signal (GHashTable *, gint );
-
 #ifndef HAVE_SETPROCTITLE
+
 /*
  * Process title utility functions
  */
-gint init_title (struct rspamd_main *, gint argc, gchar *argv[], gchar *envp[]);
+gint init_title (rspamd_mempool_t *pool, gint argc, gchar *argv[], gchar *envp[]);
+
 gint setproctitle (const gchar *fmt, ...);
+
 #endif
 
 #ifndef HAVE_PIDFILE
@@ -137,12 +138,17 @@ typedef struct rspamd_pidfh_s {
 	dev_t pf_dev;
 	ino_t pf_ino;
 } rspamd_pidfh_t;
-rspamd_pidfh_t * rspamd_pidfile_open (const gchar *path,
-	mode_t mode,
-	pid_t *pidptr);
+
+rspamd_pidfh_t *rspamd_pidfile_open (const gchar *path,
+									 mode_t mode,
+									 pid_t *pidptr);
+
 gint rspamd_pidfile_write (rspamd_pidfh_t *pfh);
+
 gint rspamd_pidfile_close (rspamd_pidfh_t *pfh);
+
 gint rspamd_pidfile_remove (rspamd_pidfh_t *pfh);
+
 #else
 typedef struct pidfh rspamd_pidfh_t;
 #define rspamd_pidfile_open pidfile_open
@@ -154,10 +160,10 @@ typedef struct pidfh rspamd_pidfh_t;
 /*
  * Replace %r with rcpt value and %f with from value, new string is allocated in pool
  */
-gchar * resolve_stat_filename (rspamd_mempool_t *pool,
-	gchar *pattern,
-	gchar *rcpt,
-	gchar *from);
+gchar *resolve_stat_filename (rspamd_mempool_t *pool,
+							  gchar *pattern,
+							  gchar *rcpt,
+							  gchar *from);
 
 const gchar *
 rspamd_log_check_time (gdouble start, gdouble end, gint resolution);
@@ -166,13 +172,8 @@ rspamd_log_check_time (gdouble start, gdouble end, gint resolution);
  * File locking functions
  */
 gboolean rspamd_file_lock (gint fd, gboolean async);
-gboolean rspamd_file_unlock (gint fd, gboolean async);
 
-/*
- * Google perf-tools initialization function
- */
-void gperf_profiler_init (struct rspamd_config *cfg, const gchar *descr);
-void gperf_profiler_stop (void);
+gboolean rspamd_file_unlock (gint fd, gboolean async);
 
 /*
  * Workarounds for older versions of glib
@@ -195,23 +196,26 @@ void g_ptr_array_insert (GPtrArray *array, gint index_, gpointer data);
 GPtrArray* g_ptr_array_new_full (guint reserved_size,
 		GDestroyNotify element_free_func);
 #endif
+#if ((GLIB_MAJOR_VERSION == 2) && (GLIB_MINOR_VERSION < 32))
+const gchar *g_environ_getenv (gchar **envp, const gchar *variable);
+#endif
 
 /*
  * Convert milliseconds to timeval fields
  */
 #define msec_to_tv(msec, tv) do { (tv)->tv_sec = (msec) / 1000; (tv)->tv_usec = \
-									  ((msec) - (tv)->tv_sec * 1000) * 1000; \
+                                      ((msec) - (tv)->tv_sec * 1000) * 1000; \
 } while (0)
 #define double_to_tv(dbl, tv) do { (tv)->tv_sec = (int)(dbl); (tv)->tv_usec = \
-									   ((dbl) - (int)(dbl)) * 1000 * 1000; \
+                                       ((dbl) - (int)(dbl)) * 1000 * 1000; \
 } while (0)
 #define double_to_ts(dbl, ts) do { (ts)->tv_sec = (int)(dbl); (ts)->tv_nsec = \
                                        ((dbl) - (int)(dbl)) * 1e9; \
 } while (0)
 #define tv_to_msec(tv) ((tv)->tv_sec * 1000LLU + (tv)->tv_usec / 1000LLU)
 #define tv_to_double(tv) ((double)(tv)->tv_sec + (tv)->tv_usec / 1.0e6)
-#define ts_to_usec(ts) ((ts)->tv_sec * 1000000LLU +							\
-	(ts)->tv_nsec / 1000LLU)
+#define ts_to_usec(ts) ((ts)->tv_sec * 1000000LLU +                            \
+    (ts)->tv_nsec / 1000LLU)
 #define ts_to_double(tv) ((double)(tv)->tv_sec + (tv)->tv_nsec / 1.0e9)
 
 /**
@@ -239,7 +243,7 @@ typedef struct rspamd_mutex_s {
  * Create new mutex
  * @return mutex or NULL
  */
-rspamd_mutex_t * rspamd_mutex_new (void);
+rspamd_mutex_t *rspamd_mutex_new (void);
 
 /**
  * Lock mutex
@@ -259,19 +263,6 @@ void rspamd_mutex_unlock (rspamd_mutex_t *mtx);
  */
 void rspamd_mutex_free (rspamd_mutex_t *mtx);
 
-/**
- * Create new named thread
- * @param name name pattern
- * @param func function to start
- * @param data data to pass to function
- * @param err error pointer
- * @return new thread object that can be joined
- */
-GThread * rspamd_create_thread (const gchar *name,
-	GThreadFunc func,
-	gpointer data,
-	GError **err);
-
 /**
  * Deep copy of one hash table to another
  * @param src source hash
@@ -281,9 +272,9 @@ GThread * rspamd_create_thread (const gchar *name,
  * @param ud user data for copy functions
  */
 void rspamd_hash_table_copy (GHashTable *src, GHashTable *dst,
-	gpointer (*key_copy_func)(gconstpointer data, gpointer ud),
-	gpointer (*value_copy_func)(gconstpointer data, gpointer ud),
-	gpointer ud);
+							 gpointer (*key_copy_func) (gconstpointer data, gpointer ud),
+							 gpointer (*value_copy_func) (gconstpointer data, gpointer ud),
+							 gpointer ud);
 
 
 /**
@@ -325,6 +316,7 @@ void rspamd_ptr_array_free_hard (gpointer p);
  * @param p
  */
 void rspamd_array_free_hard (gpointer p);
+
 /**
  * Special utility to help GString freeing in rspamd_mempool
  * @param p
@@ -343,33 +335,6 @@ void rspamd_gerror_free_maybe (gpointer p);
  */
 void rspamd_gstring_free_soft (gpointer p);
 
-struct rspamd_external_libs_ctx;
-/**
- * Initialize rspamd libraries
- */
-struct rspamd_external_libs_ctx* rspamd_init_libs (void);
-
-/**
- * Configure libraries
- */
-void rspamd_config_libs (struct rspamd_external_libs_ctx *ctx,
-		struct rspamd_config *cfg);
-
-/**
- * Reset and initialize decompressor
- * @param ctx
- */
-gboolean rspamd_libs_reset_decompression (struct rspamd_external_libs_ctx *ctx);
-/**
- * Reset and initialize compressor
- * @param ctx
- */
-gboolean rspamd_libs_reset_compression (struct rspamd_external_libs_ctx *ctx);
-
-/**
- * Destroy external libraries context
- */
-void rspamd_deinit_libs (struct rspamd_external_libs_ctx *ctx);
 
 /**
  * Returns some statically initialized random hash seed
@@ -407,7 +372,7 @@ gdouble rspamd_random_double (void);
  * @return
  */
 gdouble rspamd_random_double_fast (void);
-
+gdouble rspamd_random_double_fast_seed (guint64 seed[4]);
 guint64 rspamd_random_uint64_fast (void);
 
 /**
@@ -418,20 +383,7 @@ void rspamd_random_seed_fast (void);
 /**
  * Constant time version of memcmp
  */
-gboolean rspamd_constant_memcmp (const guchar *a, const guchar *b, gsize len);
-
-/* Special case for ancient libevent */
-#if !defined(LIBEVENT_VERSION_NUMBER) || LIBEVENT_VERSION_NUMBER < 0x02000000UL
-struct event_base * event_get_base (struct event *ev);
-#endif
-/* CentOS libevent */
-#ifndef evsignal_set
-#define evsignal_set(ev, x, cb, arg)    \
-    event_set((ev), (x), EV_SIGNAL|EV_PERSIST, (cb), (arg))
-#endif
-
-/* Avoid stupidity in libevent > 1.4 */
-int rspamd_event_pending (struct event *ev, short what);
+gboolean rspamd_constant_memcmp (const void *a, const void *b, gsize len);
 
 /**
  * Open file without following symlinks or special stuff
@@ -441,7 +393,7 @@ int rspamd_event_pending (struct event *ev, short what);
  * @return fd or -1 in case of error
  */
 int rspamd_file_xopen (const char *fname, int oflags, guint mode,
-		gboolean allow_symlink);
+					   gboolean allow_symlink);
 
 /**
  * Map file without following symlinks or special stuff
@@ -451,7 +403,7 @@ int rspamd_file_xopen (const char *fname, int oflags, guint mode,
  * @return pointer to memory (should be freed using munmap) or NULL in case of error
  */
 gpointer rspamd_file_xmap (const char *fname, guint mode, gsize *size,
-		gboolean allow_symlink);
+						   gboolean allow_symlink);
 
 /**
  * Map named shared memory segment
@@ -461,7 +413,7 @@ gpointer rspamd_file_xmap (const char *fname, guint mode, gsize *size,
  * @return pointer to memory (should be freed using munmap) or NULL in case of error
  */
 gpointer rspamd_shmem_xmap (const char *fname, guint mode,
-		gsize *size);
+							gsize *size);
 
 /**
  * Normalize probabilities using polynomial function
@@ -492,7 +444,7 @@ void rspamd_gmtime (gint64 ts, struct tm *dest);
  */
 void rspamd_localtime (gint64 ts, struct tm *dest);
 
-#define PTR_ARRAY_FOREACH(ar, i, cur) for ((i) = 0; (ar) != NULL && (i) < (ar)->len && (((cur) = g_ptr_array_index((ar), (i))) || 1); ++(i))
+#define PTR_ARRAY_FOREACH(ar, i, cur) for ((i) = 0; (ar) != NULL && (i) < (ar)->len && (((cur) = (__typeof__(cur))g_ptr_array_index((ar), (i))) || 1); ++(i))
 
 /**
  * Compresses the input string using gzip+zlib. Old string is replaced and freed
@@ -516,8 +468,8 @@ GPtrArray *rspamd_glob_path (const gchar *dir,
 							 GError **err);
 
 struct rspamd_counter_data {
-	gdouble mean;
-	gdouble stddev;
+	float mean;
+	float stddev;
 	guint64 number;
 };
 
@@ -528,9 +480,9 @@ struct rspamd_counter_data {
  * @param alpha decay coefficient (0..1)
  * @return new counter value
  */
-double rspamd_set_counter_ema (struct rspamd_counter_data *cd,
-							   gdouble value,
-							   gdouble alpha);
+float rspamd_set_counter_ema (struct rspamd_counter_data *cd,
+							   float value,
+							   float alpha);
 
 /**
  * Sets counter's data using flat moving average
@@ -541,4 +493,33 @@ double rspamd_set_counter_ema (struct rspamd_counter_data *cd,
 double rspamd_set_counter (struct rspamd_counter_data *cd,
 						   gdouble value);
 
+/**
+ * Shuffle elements in an array inplace
+ * @param ar
+ */
+void rspamd_ptr_array_shuffle (GPtrArray *ar);
+
+enum rspamd_pbkdf_version_id {
+	RSPAMD_PBKDF_ID_V1 = 1,
+	RSPAMD_PBKDF_ID_V2 = 2,
+	RSPAMD_PBKDF_ID_MAX
+};
+
+struct rspamd_controller_pbkdf {
+	const char *name;
+	const char *alias;
+	const char *description;
+	int type; /* enum rspamd_cryptobox_pbkdf_type */
+	gint id;
+	guint complexity;
+	gsize salt_len;
+	gsize key_len;
+};
+
+extern const struct rspamd_controller_pbkdf pbkdf_list[];
+
+#ifdef  __cplusplus
+}
+#endif
+
 #endif
