@@ -21,18 +21,18 @@
 #include "libutil/fstring.h"
 #include "libutil/mem_pool.h"
 #include "libutil/util.h"
-#include "libutil/logger.h"
-#include "libutil/http.h"
+#include "libserver/logger.h"
+#include "libserver/http/http_connection.h"
 #include "libutil/upstream.h"
 #include "libutil/radix.h"
+#include "libserver/cfg_file.h"
 #include "libserver/url.h"
 #include "libserver/protocol.h"
-#include "libserver/events.h"
+#include "libserver/async_session.h"
 #include "libserver/roll_history.h"
 #include "libserver/task.h"
-#include <openssl/ssl.h>
-#include <magic.h>
 
+#include <openssl/ssl.h>
 
 /* Default values */
 #define FIXED_CONFIG_FILE RSPAMD_CONFDIR "/rspamd.conf"
@@ -52,6 +52,12 @@
 #define CR '\r'
 #define LF '\n'
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
+struct rspamd_main;
+
 enum rspamd_worker_flags {
 	RSPAMD_WORKER_HAS_SOCKET = (1 << 0),
 	RSPAMD_WORKER_UNIQUE = (1 << 1),
@@ -60,8 +66,34 @@ enum rspamd_worker_flags {
 	RSPAMD_WORKER_ALWAYS_START = (1 << 4),
 	RSPAMD_WORKER_SCANNER = (1 << 5),
 	RSPAMD_WORKER_CONTROLLER = (1 << 6),
+	RSPAMD_WORKER_NO_TERMINATE_DELAY = (1 << 7),
+	RSPAMD_WORKER_OLD_CONFIG = (1 << 8),
 };
 
+struct rspamd_worker_accept_event {
+	ev_io accept_ev;
+	ev_timer throttling_ev;
+	struct ev_loop *event_loop;
+	struct rspamd_worker_accept_event *prev, *next;
+};
+
+typedef void (*rspamd_worker_term_cb) (EV_P_ ev_child *, struct rspamd_main *,
+									   struct rspamd_worker *);
+
+struct rspamd_worker_heartbeat {
+	ev_timer heartbeat_ev;          /**< used by main for checking heartbeats and by workers to send heartbeats */
+	ev_tstamp last_event;           /**< last heartbeat received timestamp */
+	gint64 nbeats;                  /**< positive for beats received, negative for beats missed */
+};
+
+enum rspamd_worker_state {
+	rspamd_worker_state_running = 0,
+	rspamd_worker_state_wanna_die,
+	rspamd_worker_state_terminating,
+	rspamd_worker_wait_connections,
+	rspamd_worker_wait_final_scripts,
+	rspamd_worker_wanna_die
+};
 
 /**
  * Worker process structure
@@ -71,30 +103,33 @@ struct rspamd_worker {
 	pid_t ppid;                     /**< pid of parent									*/
 	guint index;                    /**< index number									*/
 	guint nconns;                   /**< current connections count						*/
-	gboolean wanna_die;             /**< worker is terminating							*/
+	enum rspamd_worker_state state; /**< current worker state							*/
 	gboolean cores_throttled;       /**< set to true if cores throttling took place		*/
 	gdouble start_time;             /**< start time										*/
 	struct rspamd_main *srv;        /**< pointer to server structure					*/
 	GQuark type;                    /**< process type									*/
 	GHashTable *signal_events;      /**< signal events									*/
-	GList *accept_events;           /**< socket events									*/
+	struct rspamd_worker_accept_event *accept_events; /**< socket events				*/
 	struct rspamd_worker_conf *cf;  /**< worker config data								*/
 	gpointer ctx;                   /**< worker's specific data							*/
-	enum rspamd_worker_flags flags; /**< worker's flags									*/
+	gint flags;                     /**< worker's flags (enum rspamd_worker_flags)		*/
 	gint control_pipe[2];           /**< control pipe. [0] is used by main process,
 	                                                   [1] is used by a worker			*/
 	gint srv_pipe[2];               /**< used by workers to request something from the
 	                                     main process. [0] - main, [1] - worker			*/
-	struct event srv_ev;            /**< used by main for read workers' requests		*/
+	ev_io srv_ev;                   /**< used by main for read workers' requests		*/
+	struct rspamd_worker_heartbeat hb; /**< heartbeat data */
 	gpointer control_data;          /**< used by control protocol to handle commands	*/
 	gpointer tmp_data;              /**< used to avoid race condition to deal with control messages */
-	GPtrArray *finish_actions;      /**< called when worker is terminated				*/
+	ev_child cld_ev;                /**< to allow reaping								*/
+	rspamd_worker_term_cb term_handler; /**< custom term handler						*/
+	GHashTable *control_events_pending; /**< control events pending indexed by ptr		*/
 };
 
 struct rspamd_abstract_worker_ctx {
 	guint64 magic;
 	/* Events base */
-	struct event_base *ev_base;
+	struct ev_loop *event_loop;
 	/* DNS resolver */
 	struct rspamd_dns_resolver *resolver;
 	/* Config */
@@ -103,43 +138,34 @@ struct rspamd_abstract_worker_ctx {
 };
 
 struct rspamd_worker_signal_handler;
-typedef gboolean (*rspamd_worker_signal_handler) (
+
+typedef gboolean (*rspamd_worker_signal_cb_t) (
 		struct rspamd_worker_signal_handler *, void *ud);
 
-struct rspamd_worker_signal_cb {
-	rspamd_worker_signal_handler handler;
+struct rspamd_worker_signal_handler_elt {
+	rspamd_worker_signal_cb_t handler;
 	void *handler_data;
-	struct rspamd_worker_signal_cb *next, *prev;
+	struct rspamd_worker_signal_handler_elt *next, *prev;
 };
 
 struct rspamd_worker_signal_handler {
 	gint signo;
 	gboolean enabled;
-	struct event ev;
-	struct event_base *base;
+	ev_signal ev_sig;
+	struct ev_loop *event_loop;
 	struct rspamd_worker *worker;
-	struct rspamd_worker_signal_cb *cb;
-};
-
-struct rspamd_controller_pbkdf {
-	const char *name;
-	const char *alias;
-	const char *description;
-	enum rspamd_cryptobox_pbkdf_type type;
-	gint id;
-	guint complexity;
-	gsize salt_len;
-	gsize key_len;
+	struct rspamd_worker_signal_handler_elt *cb;
 };
 
 /**
  * Common structure representing C module context
  */
 struct module_s;
+
 struct module_ctx {
-	gint (*filter)(struct rspamd_task *task);                   /**< pointer to headers process function			*/
-	struct module_s *mod;										/**< module pointer									*/
-	gboolean enabled;											/**< true if module is enabled in configuration		*/
+	gint (*filter) (struct rspamd_task *task);                   /**< pointer to headers process function			*/
+	struct module_s *mod;                                        /**< module pointer									*/
+	gboolean enabled;                                            /**< true if module is enabled in configuration		*/
 };
 
 #ifndef WITH_HYPERSCAN
@@ -167,28 +193,33 @@ struct module_ctx {
 #define RSPAMD_CUR_WORKER_VERSION 0x2
 
 #define RSPAMD_FEATURES \
-		RSPAMD_FEATURE_HYPERSCAN RSPAMD_FEATURE_PCRE2 \
-		RSPAMD_FEATURE_FANN RSPAMD_FEATURE_SNOWBALL
+        RSPAMD_FEATURE_HYPERSCAN RSPAMD_FEATURE_PCRE2 \
+        RSPAMD_FEATURE_FANN RSPAMD_FEATURE_SNOWBALL
 
 #define RSPAMD_MODULE_VER \
-		RSPAMD_CUR_MODULE_VERSION, /* Module version */ \
-		RSPAMD_VERSION_NUM, /* Rspamd version */ \
-		RSPAMD_FEATURES /* Compilation features */ \
+        RSPAMD_CUR_MODULE_VERSION, /* Module version */ \
+        RSPAMD_VERSION_NUM, /* Rspamd version */ \
+        RSPAMD_FEATURES /* Compilation features */ \
 
 #define RSPAMD_WORKER_VER \
-		RSPAMD_CUR_WORKER_VERSION, /* Worker version */ \
-		RSPAMD_VERSION_NUM, /* Rspamd version */ \
-		RSPAMD_FEATURES /* Compilation features */ \
+        RSPAMD_CUR_WORKER_VERSION, /* Worker version */ \
+        RSPAMD_VERSION_NUM, /* Rspamd version */ \
+        RSPAMD_FEATURES /* Compilation features */ \
 /**
  * Module
  */
 typedef struct module_s {
 	const gchar *name;
-	int (*module_init_func)(struct rspamd_config *cfg, struct module_ctx **ctx);
-	int (*module_config_func)(struct rspamd_config *cfg);
-	int (*module_reconfig_func)(struct rspamd_config *cfg);
-	int (*module_attach_controller_func)(struct module_ctx *ctx,
-		GHashTable *custom_commands);
+
+	int (*module_init_func) (struct rspamd_config *cfg, struct module_ctx **ctx);
+
+	int (*module_config_func) (struct rspamd_config *cfg, bool validate);
+
+	int (*module_reconfig_func) (struct rspamd_config *cfg);
+
+	int (*module_attach_controller_func) (struct module_ctx *ctx,
+										  GHashTable *custom_commands);
+
 	guint module_version;
 	guint64 rspamd_version;
 	const gchar *rspamd_features;
@@ -205,14 +236,18 @@ struct rspamd_worker_listen_socket {
 	const rspamd_inet_addr_t *addr;
 	gint fd;
 	enum rspamd_worker_socket_type type;
+	bool is_systemd;
 };
 
 typedef struct worker_s {
 	const gchar *name;
-	gpointer (*worker_init_func)(struct rspamd_config *cfg);
-	void (*worker_start_func)(struct rspamd_worker *worker);
-	enum rspamd_worker_flags flags;
-	enum rspamd_worker_socket_type listen_type;
+
+	gpointer (*worker_init_func) (struct rspamd_config *cfg);
+
+	void (*worker_start_func) (struct rspamd_worker *worker);
+
+	int flags;
+	int listen_type;
 	guint worker_version;
 	guint64 rspamd_version;
 	const gchar *rspamd_features;
@@ -274,22 +309,12 @@ struct rspamd_main {
 	uid_t workers_uid;                                          /**< worker's uid running to                        */
 	gid_t workers_gid;                                          /**< worker's gid running to						*/
 	gboolean is_privilleged;                                    /**< true if run in privilleged mode                */
+	gboolean wanna_die;                                         /**< no respawn of processes						*/
 	gboolean cores_throttling;                                  /**< turn off cores when limits are exceeded		*/
 	struct roll_history *history;                               /**< rolling history								*/
-	struct event_base *ev_base;
-};
-
-enum rspamd_exception_type {
-	RSPAMD_EXCEPTION_NEWLINE = 0,
-	RSPAMD_EXCEPTION_URL,
-};
-/**
- * Structure to point exception in text from processing
- */
-struct rspamd_process_exception {
-	goffset pos;
-	guint len;
-	enum rspamd_exception_type type;
+	struct ev_loop *event_loop;
+	ev_signal term_ev, int_ev, hup_ev, usr1_ev;                 /**< signals 										*/
+	struct rspamd_http_context *http_ctx;
 };
 
 /**
@@ -297,8 +322,9 @@ struct rspamd_process_exception {
  */
 struct controller_command;
 struct controller_session;
-typedef gboolean (*controller_func_t)(gchar **args,
-	struct controller_session *session);
+
+typedef gboolean (*controller_func_t) (gchar **args,
+									   struct controller_session *session);
 
 struct controller_session {
 	struct rspamd_worker *worker;                               /**< pointer to worker structure (controller in fact) */
@@ -306,9 +332,9 @@ struct controller_session {
 	struct controller_command *cmd;                             /**< real command									*/
 	struct rspamd_config *cfg;                                  /**< pointer to config file							*/
 	GList *parts;                                               /**< extracted mime parts							*/
-	struct rspamd_async_session * s;                             /**< async session object							*/
+	struct rspamd_async_session *s;                             /**< async session object							*/
 	struct rspamd_dns_resolver *resolver;                       /**< DNS resolver									*/
-	struct event_base *ev_base;                                 /**< Event base										*/
+	struct ev_loop *ev_base;                                 /**< Event base										*/
 };
 
 struct zstd_dictionary {
@@ -317,11 +343,8 @@ struct zstd_dictionary {
 	guint id;
 };
 
-struct rspamd_radix_map_helper;
-
 struct rspamd_external_libs_ctx {
-	magic_t libmagic;
-	struct rspamd_radix_map_helper **local_addrs;
+	void **local_addrs;
 	struct rspamd_cryptobox_library_ctx *crypto_ctx;
 	struct ottery_config *ottery_cfg;
 	SSL_CTX *ssl_ctx;
@@ -338,17 +361,12 @@ struct rspamd_external_libs_ctx {
  * Register custom controller function
  */
 void register_custom_controller_command (const gchar *name,
-	controller_func_t handler,
-	gboolean privilleged,
-	gboolean require_message);
-
-enum rspamd_pbkdf_version_id {
-	RSPAMD_PBKDF_ID_V1 = 1,
-	RSPAMD_PBKDF_ID_V2= 2,
-	RSPAMD_PBKDF_ID_MAX
-};
-
-extern const struct rspamd_controller_pbkdf pbkdf_list[];
+										 controller_func_t handler,
+										 gboolean privilleged,
+										 gboolean require_message);
 
+#ifdef  __cplusplus
+}
+#endif
 
 #endif
\ No newline at end of file
