@@ -15,10 +15,10 @@
  */
 #include "config.h"
 #include "libutil/util.h"
-#include "libutil/map.h"
+#include "libserver/maps/map.h"
 #include "libutil/upstream.h"
-#include "libutil/http.h"
-#include "libutil/http_private.h"
+#include "libserver/http/http_connection.h"
+#include "libserver/http/http_private.h"
 #include "libserver/protocol.h"
 #include "libserver/protocol_internal.h"
 #include "libserver/cfg_file.h"
@@ -36,12 +36,19 @@
 #include "libserver/milter.h"
 #include "libserver/milter_internal.h"
 #include "libmime/lang_detection.h"
-#include "contrib/zstd/zstd.h"
+
+#include <math.h>
 
 #ifdef HAVE_NETINET_TCP_H
 #include <netinet/tcp.h> /* for TCP_NODELAY */
 #endif
 
+#ifdef SYS_ZSTD
+#  include "zstd.h"
+#else
+#  include "contrib/zstd/zstd.h"
+#endif
+
 /* Rotate keys each minute by default */
 #define DEFAULT_ROTATION_TIME 60.0
 #define DEFAULT_RETRIES 5
@@ -84,7 +91,6 @@ struct rspamd_http_upstream {
 	struct upstream_list *u;
 	struct rspamd_cryptobox_pubkey *key;
 	gdouble timeout;
-	struct timeval io_tv;
 	gint parser_from_ref;
 	gint parser_to_ref;
 	gboolean local;
@@ -99,7 +105,6 @@ struct rspamd_http_mirror {
 	struct rspamd_cryptobox_pubkey *key;
 	gdouble prob;
 	gdouble timeout;
-	struct timeval io_tv;
 	gint parser_from_ref;
 	gint parser_to_ref;
 	gboolean local;
@@ -111,28 +116,23 @@ static const guint64 rspamd_rspamd_proxy_magic = 0xcdeb4fd1fc351980ULL;
 struct rspamd_proxy_ctx {
 	guint64 magic;
 	/* Events base */
-	struct event_base *ev_base;
+	struct ev_loop *event_loop;
 	/* DNS resolver */
 	struct rspamd_dns_resolver *resolver;
 	/* Config */
 	struct rspamd_config *cfg;
 	/* END OF COMMON PART */
 	gdouble timeout;
-	struct timeval io_tv;
 	/* Encryption key for clients */
 	struct rspamd_cryptobox_keypair *key;
-	/* Keys cache */
-	struct rspamd_keypair_cache *keys_cache;
+	/* HTTP context */
+	struct rspamd_http_context *http_ctx;
 	/* Upstreams to use */
 	GHashTable *upstreams;
 	/* Mirrors to send traffic to */
 	GPtrArray *mirrors;
 	/* Default upstream */
 	struct rspamd_http_upstream *default_upstream;
-	/* Local rotating keypair for upstreams */
-	struct rspamd_cryptobox_keypair *local_key;
-	struct event rotate_ev;
-	gdouble rotate_tm;
 	lua_State *lua_state;
 	/* Array of callback functions called on end of scan to compare results */
 	GArray *cmp_refs;
@@ -176,8 +176,8 @@ struct rspamd_proxy_backend_connection {
 	ucl_object_t *results;
 	const gchar *err;
 	struct rspamd_proxy_session *s;
-	struct timeval *io_tv;
 	gint backend_sock;
+	ev_tstamp timeout;
 	enum rspamd_backend_flags flags;
 	gint parser_from_ref;
 	gint parser_to_ref;
@@ -226,7 +226,6 @@ rspamd_proxy_parse_lua_parser (lua_State *L, const ucl_object_t *obj,
 	const gchar *lua_script;
 	gsize slen;
 	gint err_idx, ref_idx;
-	GString *tb = NULL;
 	gboolean has_ref = FALSE;
 
 	g_assert (obj != NULL);
@@ -254,13 +253,11 @@ rspamd_proxy_parse_lua_parser (lua_State *L, const ucl_object_t *obj,
 
 	/* Now do it */
 	if (lua_pcall (L, 0, 1, err_idx) != 0) {
-		tb = lua_touserdata (L, -1);
 		g_set_error (err,
 				rspamd_proxy_quark (),
 				EINVAL,
 				"cannot init lua parser script: %s",
-				tb->str);
-		g_string_free (tb, TRUE);
+				lua_tostring (L, -1));
 		lua_settop (L, 0);
 
 		return FALSE;
@@ -416,6 +413,7 @@ rspamd_proxy_parse_upstream (rspamd_mempool_t *pool,
 
 	if (elt) {
 		up->u = rspamd_upstreams_create (ctx->cfg->ups_ctx);
+
 		if (!rspamd_upstreams_from_ucl (up->u, elt, 11333, NULL)) {
 			g_set_error (err, rspamd_proxy_quark (), 100,
 					"upstream has bad hosts definition");
@@ -468,8 +466,6 @@ rspamd_proxy_parse_upstream (rspamd_mempool_t *pool,
 		rspamd_lua_add_ref_dtor (L, pool, up->parser_to_ref);
 	}
 
-	double_to_tv (up->timeout, &up->io_tv);
-
 	g_hash_table_insert (ctx->upstreams, up->name, up);
 
 	return TRUE;
@@ -621,8 +617,6 @@ rspamd_proxy_parse_mirror (rspamd_mempool_t *pool,
 		up->settings_id = rspamd_mempool_strdup (pool, ucl_object_tostring (elt));
 	}
 
-	double_to_tv (up->timeout, &up->io_tv);
-
 	g_ptr_array_add (ctx->mirrors, up);
 
 	return TRUE;
@@ -645,7 +639,6 @@ rspamd_proxy_parse_script (rspamd_mempool_t *pool,
 	const gchar *lua_script;
 	gsize slen;
 	gint err_idx, ref_idx;
-	GString *tb = NULL;
 	struct stat st;
 
 	ctx = pd->user_struct;
@@ -691,13 +684,11 @@ rspamd_proxy_parse_script (rspamd_mempool_t *pool,
 
 	/* Now do it */
 	if (lua_pcall (L, 0, 1, err_idx) != 0) {
-		tb = lua_touserdata (L, -1);
 		g_set_error (err,
 				rspamd_proxy_quark (),
 				EINVAL,
 				"cannot init lua parser script: %s",
-				tb->str);
-		g_string_free (tb, TRUE);
+				lua_tostring (L, -1));
 		lua_settop (L, 0);
 
 		goto err;
@@ -743,7 +734,6 @@ init_rspamd_proxy (struct rspamd_config *cfg)
 	ctx->mirrors = g_ptr_array_new ();
 	rspamd_mempool_add_destructor (cfg->cfg_pool,
 			(rspamd_mempool_destruct_t)rspamd_ptr_array_free_hard, ctx->mirrors);
-	ctx->rotate_tm = DEFAULT_ROTATION_TIME;
 	ctx->cfg = cfg;
 	ctx->lua_state = cfg->lua_state;
 	ctx->cmp_refs = g_array_new (FALSE, FALSE, sizeof (gint));
@@ -760,15 +750,6 @@ init_rspamd_proxy (struct rspamd_config *cfg)
 			G_STRUCT_OFFSET (struct rspamd_proxy_ctx, timeout),
 			RSPAMD_CL_FLAG_TIME_FLOAT,
 			"IO timeout");
-	rspamd_rcl_register_worker_option (cfg,
-			type,
-			"rotate",
-			rspamd_rcl_parse_struct_time,
-			ctx,
-			G_STRUCT_OFFSET (struct rspamd_proxy_ctx, rotate_tm),
-			RSPAMD_CL_FLAG_TIME_FLOAT,
-			"Rotation keys time, default: "
-			G_STRINGIFY (DEFAULT_ROTATION_TIME) " seconds");
 	rspamd_rcl_register_worker_option (cfg,
 			type,
 			"keypair",
@@ -878,18 +859,37 @@ proxy_backend_close_connection (struct rspamd_proxy_backend_connection *conn)
 
 static gboolean
 proxy_backend_parse_results (struct rspamd_proxy_session *session,
-		struct rspamd_proxy_backend_connection *conn,
-		lua_State *L, gint parser_ref,
-		const gchar *in, gsize inlen)
+							 struct rspamd_proxy_backend_connection *conn,
+							 lua_State *L, gint parser_ref,
+							 struct rspamd_http_message *msg,
+							 goffset *body_offset,
+							 const rspamd_ftok_t *ct)
 {
 	struct ucl_parser *parser;
-	GString *tb = NULL;
 	gint err_idx;
+	const gchar *in = msg->body_buf.begin;
+	gsize inlen = msg->body_buf.len;
+	const rspamd_ftok_t *offset_hdr;
 
 	if (inlen == 0 || in == NULL) {
 		return FALSE;
 	}
 
+	offset_hdr = rspamd_http_message_find_header (msg, MESSAGE_OFFSET_HEADER);
+
+	if (offset_hdr) {
+		gulong val;
+
+		if (rspamd_strtoul (offset_hdr->begin, offset_hdr->len, &val)
+			&& val < inlen) {
+
+			if (body_offset) {
+				*body_offset = val;
+			}
+			inlen = val;
+		}
+	}
+
 	if (parser_ref != -1) {
 		/* Call parser function */
 		lua_pushcfunction (L, &rspamd_lua_traceback);
@@ -900,11 +900,9 @@ proxy_backend_parse_results (struct rspamd_proxy_session *session,
 		lua_pushlstring (L, in, inlen);
 
 		if (lua_pcall (L, 1, 1, err_idx) != 0) {
-			tb = lua_touserdata (L, -1);
 			msg_err_session (
 					"cannot run lua parser script: %s",
-					tb->str);
-			g_string_free (tb, TRUE);
+					lua_tostring (L, -1));
 			lua_settop (L, 0);
 
 			return FALSE;
@@ -914,23 +912,28 @@ proxy_backend_parse_results (struct rspamd_proxy_session *session,
 		lua_settop (L, 0);
 	}
 	else {
-		parser = ucl_parser_new (0);
+		rspamd_ftok_t json_ct;
+		RSPAMD_FTOK_ASSIGN (&json_ct, "application/json");
 
-		if (!ucl_parser_add_chunk (parser, in, inlen)) {
-			gchar *encoded;
+		if (ct && rspamd_ftok_casecmp (ct, &json_ct) == 0) {
+			parser = ucl_parser_new (0);
 
-			encoded = rspamd_encode_base64 (in, inlen, 0, NULL);
-			msg_err_session ("cannot parse input: %s", ucl_parser_get_error (
-					parser));
-			msg_err_session ("input encoded: %s", encoded);
+			if (!ucl_parser_add_chunk (parser, in, inlen)) {
+				gchar *encoded;
+
+				encoded = rspamd_encode_base64 (in, inlen, 0, NULL);
+				msg_err_session ("cannot parse input: %s", ucl_parser_get_error (
+						parser));
+				msg_err_session ("input encoded: %s", encoded);
+				ucl_parser_free (parser);
+				g_free (encoded);
+
+				return FALSE;
+			}
+
+			conn->results = ucl_parser_get_object (parser);
 			ucl_parser_free (parser);
-			g_free (encoded);
-
-			return FALSE;
 		}
-
-		conn->results = ucl_parser_get_object (parser);
-		ucl_parser_free (parser);
 	}
 
 	return TRUE;
@@ -939,7 +942,6 @@ proxy_backend_parse_results (struct rspamd_proxy_session *session,
 static void
 proxy_call_cmp_script (struct rspamd_proxy_session *session, gint cbref)
 {
-	GString *tb = NULL;
 	gint err_idx;
 	guint i;
 	struct rspamd_proxy_backend_connection *conn;
@@ -980,11 +982,9 @@ proxy_call_cmp_script (struct rspamd_proxy_session *session, gint cbref)
 	}
 
 	if (lua_pcall (L, 1, 0, err_idx) != 0) {
-		tb = lua_touserdata (L, -1);
 		msg_err_session (
 				"cannot run lua compare script: %s",
-				tb->str);
-		g_string_free (tb, TRUE);
+				lua_tostring (L, -1));
 	}
 
 	lua_settop (L, 0);
@@ -1077,7 +1077,7 @@ proxy_request_compress (struct rspamd_http_message *msg)
 
 	flags = rspamd_http_message_get_flags (msg);
 
-	if (!rspamd_http_message_find_header (msg, "Compression")) {
+	if (!rspamd_http_message_find_header (msg, COMPRESSION_HEADER)) {
 		if ((flags & RSPAMD_HTTP_FLAG_SHMEM) ||
 				!(flags & RSPAMD_HTTP_FLAG_HAS_BODY)) {
 			/* Cannot compress shared or empty message */
@@ -1105,7 +1105,7 @@ proxy_request_compress (struct rspamd_http_message *msg)
 
 		ZSTD_freeCCtx (zctx);
 		rspamd_http_message_set_body_from_fstring_steal (msg, body);
-		rspamd_http_message_add_header (msg, "Compression", "zstd");
+		rspamd_http_message_add_header (msg, COMPRESSION_HEADER, "zstd");
 	}
 }
 
@@ -1119,7 +1119,7 @@ proxy_request_decompress (struct rspamd_http_message *msg)
 	ZSTD_inBuffer zin;
 	ZSTD_outBuffer zout;
 
-	if (rspamd_http_message_find_header (msg, "Compression")) {
+	if (rspamd_http_message_find_header (msg, COMPRESSION_HEADER)) {
 		in = rspamd_http_message_get_body (msg, &inlen);
 
 		if (in == NULL || inlen == 0) {
@@ -1155,7 +1155,7 @@ proxy_request_decompress (struct rspamd_http_message *msg)
 
 			if (zout.pos == zout.size) {
 				/* We need to extend output buffer */
-				zout.size = zout.size * 1.5 + 1.0;
+				zout.size = zout.size * 2 + 1;
 				body = rspamd_fstring_grow (body, zout.size);
 				zout.size = body->allocated;
 				zout.dst = body->str;
@@ -1165,10 +1165,8 @@ proxy_request_decompress (struct rspamd_http_message *msg)
 		body->len = zout.pos;
 		ZSTD_freeDStream (zstream);
 		rspamd_http_message_set_body_from_fstring_steal (msg, body);
-		rspamd_http_message_remove_header (msg, "Compression");
+		rspamd_http_message_remove_header (msg, COMPRESSION_HEADER);
 	}
-
-	return;
 }
 
 static struct rspamd_proxy_session *
@@ -1185,7 +1183,7 @@ proxy_session_refresh (struct rspamd_proxy_session *session)
 	session->client_addr = NULL;
 	nsession->ctx = session->ctx;
 	nsession->worker = session->worker;
-	nsession->pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), "proxy");
+	nsession->pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), "proxy", 0);
 	nsession->client_sock = session->client_sock;
 	session->client_sock = -1;
 	nsession->mirror_conns = g_ptr_array_sized_new (nsession->ctx->mirrors->len);
@@ -1306,14 +1304,15 @@ proxy_backend_mirror_error_handler (struct rspamd_http_connection *conn, GError
 	msg_info_session ("abnormally closing connection from backend: %s:%s, "
 			"error: %e",
 			bk_conn->name,
-			rspamd_inet_address_to_string (rspamd_upstream_addr (bk_conn->up)),
+			rspamd_inet_address_to_string_pretty (
+					rspamd_upstream_addr_cur (bk_conn->up)),
 			err);
 
 	if (err) {
 		bk_conn->err = rspamd_mempool_strdup (session->pool, err->message);
 	}
 
-	rspamd_upstream_fail (bk_conn->up, FALSE);
+	rspamd_upstream_fail (bk_conn->up, FALSE, err ? err->message : "unknown");
 
 	proxy_backend_close_connection (bk_conn);
 	REF_RELEASE (bk_conn->s);
@@ -1325,16 +1324,19 @@ proxy_backend_mirror_finish_handler (struct rspamd_http_connection *conn,
 {
 	struct rspamd_proxy_backend_connection *bk_conn = conn->ud;
 	struct rspamd_proxy_session *session;
+	const rspamd_ftok_t *orig_ct;
 
 	session = bk_conn->s;
 
 	proxy_request_decompress (msg);
+	orig_ct = rspamd_http_message_find_header (msg, "Content-Type");
 
 	if (!proxy_backend_parse_results (session, bk_conn, session->ctx->lua_state,
-			bk_conn->parser_from_ref, msg->body_buf.begin, msg->body_buf.len)) {
+			bk_conn->parser_from_ref, msg, NULL, orig_ct)) {
 		msg_warn_session ("cannot parse results from the mirror backend %s:%s",
 				bk_conn->name,
-				rspamd_inet_address_to_string (rspamd_upstream_addr (bk_conn->up)));
+				rspamd_inet_address_to_string (
+						rspamd_upstream_addr_cur (bk_conn->up)));
 		bk_conn->err = "cannot parse ucl";
 	}
 
@@ -1371,7 +1373,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 				sizeof (*bk_conn));
 		bk_conn->s = session;
 		bk_conn->name = m->name;
-		bk_conn->io_tv = &m->io_tv;
+		bk_conn->timeout = m->timeout;
 
 		bk_conn->up = rspamd_upstream_get (m->u,
 				RSPAMD_UPSTREAM_ROUND_ROBIN, NULL, 0);
@@ -1384,12 +1386,12 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 		}
 
 		bk_conn->backend_sock = rspamd_inet_address_connect (
-				rspamd_upstream_addr (bk_conn->up),
+				rspamd_upstream_addr_next (bk_conn->up),
 				SOCK_STREAM, TRUE);
 
 		if (bk_conn->backend_sock == -1) {
 			msg_err_session ("cannot connect upstream for %s", m->name);
-			rspamd_upstream_fail (bk_conn->up, TRUE);
+			rspamd_upstream_fail (bk_conn->up, TRUE, strerror (errno));
 			continue;
 		}
 
@@ -1413,23 +1415,20 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 			rspamd_http_message_add_header (msg, "Settings-ID", m->settings_id);
 		}
 
-		bk_conn->backend_conn = rspamd_http_connection_new (NULL,
+		bk_conn->backend_conn = rspamd_http_connection_new_client_socket (
+				session->ctx->http_ctx,
+				NULL,
 				proxy_backend_mirror_error_handler,
 				proxy_backend_mirror_finish_handler,
 				RSPAMD_HTTP_CLIENT_SIMPLE,
-				RSPAMD_HTTP_CLIENT,
-				session->ctx->keys_cache,
-				NULL);
+				bk_conn->backend_sock);
 
 		if (m->key) {
-			rspamd_http_connection_set_key (bk_conn->backend_conn,
-					session->ctx->local_key);
 			msg->peer_key = rspamd_pubkey_ref (m->key);
 		}
 
 		if (m->local ||
-				rspamd_inet_address_is_local (
-						rspamd_upstream_addr (bk_conn->up), FALSE)) {
+				rspamd_inet_address_is_local (rspamd_upstream_addr_cur (bk_conn->up))) {
 
 			if (session->fname) {
 				rspamd_http_message_add_header (msg, "File", session->fname);
@@ -1438,8 +1437,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 			msg->method = HTTP_GET;
 			rspamd_http_connection_write_message_shared (bk_conn->backend_conn,
 					msg, NULL, NULL, bk_conn,
-					bk_conn->backend_sock,
-					bk_conn->io_tv, session->ctx->ev_base);
+					bk_conn->timeout);
 		}
 		else {
 			if (session->fname) {
@@ -1466,8 +1464,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 
 			rspamd_http_connection_write_message (bk_conn->backend_conn,
 					msg, NULL, NULL, bk_conn,
-					bk_conn->backend_sock,
-					bk_conn->io_tv, session->ctx->ev_base);
+					bk_conn->timeout);
 		}
 
 		g_ptr_array_add (session->mirror_conns, bk_conn);
@@ -1489,11 +1486,39 @@ proxy_client_write_error (struct rspamd_proxy_session *session, gint code,
 	}
 	else {
 		reply = rspamd_http_new_message (HTTP_RESPONSE);
-		reply->code = code;
-		reply->status = rspamd_fstring_new_init (status, strlen (status));
+
+		switch (code) {
+		case ETIMEDOUT:
+			reply->code = 504;
+			reply->status = RSPAMD_FSTRING_LIT ("Gateway timeout");
+			break;
+		case ECONNRESET:
+		case ECONNABORTED:
+			reply->code = 502;
+			reply->status = RSPAMD_FSTRING_LIT ("Gateway connection reset");
+			break;
+		case ECONNREFUSED:
+			reply->code = 502;
+			reply->status = RSPAMD_FSTRING_LIT ("Gateway connection refused");
+			break;
+		default:
+			if (code >= 300) {
+				/* Likely HTTP error */
+				reply->code = code;
+				reply->status = rspamd_fstring_new_init (status, strlen (status));
+			}
+			else {
+				reply->code = 502;
+				reply->status = RSPAMD_FSTRING_LIT ("Unknown gateway error: ");
+				reply->status = rspamd_fstring_append (reply->status,
+						status, strlen (status));
+			}
+			break;
+		}
+
 		rspamd_http_connection_write_message (session->client_conn,
-				reply, NULL, NULL, session, session->client_sock,
-				&session->ctx->io_tv, session->ctx->ev_base);
+				reply, NULL, NULL, session,
+				session->ctx->timeout);
 	}
 }
 
@@ -1504,31 +1529,42 @@ proxy_backend_master_error_handler (struct rspamd_http_connection *conn, GError
 	struct rspamd_proxy_session *session;
 
 	session = bk_conn->s;
-	msg_info_session ("abnormally closing connection from backend: %s, error: %e,"
-			" retries left: %d",
-		rspamd_inet_address_to_string (rspamd_upstream_addr (session->master_conn->up)),
-		err,
-		session->ctx->max_retries - session->retries);
 	session->retries ++;
-	rspamd_upstream_fail (bk_conn->up, FALSE);
+	msg_info_session ("abnormally closing connection from backend: %s, error: %e,"
+					  " retries left: %d",
+			rspamd_inet_address_to_string_pretty (
+					rspamd_upstream_addr_cur (session->master_conn->up)),
+			err,
+			session->ctx->max_retries - session->retries);
+	rspamd_upstream_fail (bk_conn->up, FALSE, err ? err->message : "unknown");
 	proxy_backend_close_connection (session->master_conn);
 
-	if (session->ctx->max_retries &&
-			session->retries > session->ctx->max_retries) {
+	if (session->ctx->max_retries > 0 &&
+			session->retries >= session->ctx->max_retries) {
 		msg_err_session ("cannot connect to upstream, maximum retries "
 				"has been reached: %d", session->retries);
 		/* Terminate session immediately */
-		proxy_client_write_error (session, err->code, err->message);
+		if (err) {
+			proxy_client_write_error(session, err->code, err->message);
+		}
+		else {
+			proxy_client_write_error(session, 503, "Unknown error after no retries left");
+		}
 	}
 	else {
 		if (!proxy_send_master_message (session)) {
-			proxy_client_write_error (session, err->code, err->message);
+			if (err) {
+				proxy_client_write_error(session, err->code, err->message);
+			}
+			else {
+				proxy_client_write_error(session, 503, "Unknown error on write");
+			}
 		}
 		else {
 			msg_info_session ("retry connection to: %s"
 					" retries left: %d",
 					rspamd_inet_address_to_string (
-							rspamd_upstream_addr (session->master_conn->up)),
+							rspamd_upstream_addr_cur (session->master_conn->up)),
 					session->ctx->max_retries - session->retries);
 		}
 	}
@@ -1536,22 +1572,32 @@ proxy_backend_master_error_handler (struct rspamd_http_connection *conn, GError
 
 static gint
 proxy_backend_master_finish_handler (struct rspamd_http_connection *conn,
-	struct rspamd_http_message *msg)
+									 struct rspamd_http_message *msg)
 {
 	struct rspamd_proxy_backend_connection *bk_conn = conn->ud;
 	struct rspamd_proxy_session *session, *nsession;
 	rspamd_fstring_t *reply;
+	const rspamd_ftok_t *orig_ct;
+	goffset body_offset = -1;
 
 	session = bk_conn->s;
 	rspamd_http_connection_steal_msg (session->master_conn->backend_conn);
 	proxy_request_decompress (msg);
 
+	/*
+	 * These are likely set by an http library, so we will double these headers
+	 * if they are not removed
+	 */
 	rspamd_http_message_remove_header (msg, "Content-Length");
+	rspamd_http_message_remove_header (msg, "Connection");
+	rspamd_http_message_remove_header (msg, "Date");
+	rspamd_http_message_remove_header (msg, "Server");
 	rspamd_http_message_remove_header (msg, "Key");
+	orig_ct = rspamd_http_message_find_header (msg, "Content-Type");
 	rspamd_http_connection_reset (session->master_conn->backend_conn);
 
 	if (!proxy_backend_parse_results (session, bk_conn, session->ctx->lua_state,
-			bk_conn->parser_from_ref, msg->body_buf.begin, msg->body_buf.len)) {
+			bk_conn->parser_from_ref, msg, &body_offset, orig_ct)) {
 		msg_warn_session ("cannot parse results from the master backend");
 	}
 
@@ -1582,15 +1628,32 @@ proxy_backend_master_finish_handler (struct rspamd_http_connection *conn,
 
 	if (session->client_milter_conn) {
 		nsession = proxy_session_refresh (session);
-		rspamd_milter_send_task_results (nsession->client_milter_conn,
-				session->master_conn->results);
+
+		if (body_offset > 0) {
+			rspamd_milter_send_task_results (nsession->client_milter_conn,
+					session->master_conn->results,
+					msg->body_buf.begin + body_offset,
+					msg->body_buf.len - body_offset);
+		}
+		else {
+			rspamd_milter_send_task_results (nsession->client_milter_conn,
+					session->master_conn->results, NULL, 0);
+		}
 		REF_RELEASE (session);
 		rspamd_http_message_free (msg);
 	}
 	else {
+		const gchar *passed_ct = NULL;
+
+		if (orig_ct) {
+			passed_ct = rspamd_mempool_ftokdup (session->pool, orig_ct);
+			/* Remove original */
+			rspamd_http_message_remove_header (msg, "Content-Type");
+		}
+
 		rspamd_http_connection_write_message (session->client_conn,
-				msg, NULL, NULL, session, session->client_sock,
-				bk_conn->io_tv, session->ctx->ev_base);
+				msg, NULL, passed_ct, session,
+				bk_conn->timeout);
 	}
 
 	return 0;
@@ -1609,12 +1672,10 @@ rspamd_proxy_scan_self_reply (struct rspamd_task *task)
 	msg->code = 200;
 
 	switch (task->cmd) {
-	case CMD_REPORT_IFSPAM:
-	case CMD_REPORT:
 	case CMD_CHECK:
-	case CMD_SYMBOLS:
-	case CMD_PROCESS:
 	case CMD_SKIP:
+	case CMD_CHECK_RSPAMC:
+	case CMD_CHECK_SPAMC:
 	case CMD_CHECK_V2:
 		rspamd_task_set_finish_time (task);
 		rspamd_protocol_http_reply (msg, task, &rep);
@@ -1624,7 +1685,7 @@ rspamd_proxy_scan_self_reply (struct rspamd_task *task)
 		rspamd_http_message_set_body (msg, "pong" CRLF, 6);
 		ctype = "text/plain";
 		break;
-	case CMD_OTHER:
+	default:
 		msg_err_task ("BROKEN");
 		break;
 	}
@@ -1637,8 +1698,50 @@ rspamd_proxy_scan_self_reply (struct rspamd_task *task)
 
 	if (session->client_milter_conn) {
 		nsession = proxy_session_refresh (session);
-		rspamd_milter_send_task_results (nsession->client_milter_conn,
-				session->master_conn->results);
+
+		if (task->flags & RSPAMD_TASK_FLAG_MESSAGE_REWRITE) {
+			const gchar *start;
+			goffset len, hdr_off;
+
+			start = task->msg.begin;
+			len = task->msg.len;
+
+			hdr_off = MESSAGE_FIELD (task, raw_headers_content).len;
+
+			if (hdr_off < len) {
+				start += hdr_off;
+				len -= hdr_off;
+
+				/* The problem here is that we need not end of headers, we need
+				 * start of body.
+				 *
+				 * Hence, we need to skip one \r\n till there is anything else in
+				 * a line.
+				 */
+
+				if (*start == '\r' && len > 0) {
+					start++;
+					len--;
+				}
+
+				if (*start == '\n' && len > 0) {
+					start++;
+					len--;
+				}
+
+				rspamd_milter_send_task_results (nsession->client_milter_conn,
+						session->master_conn->results, start, len);
+			}
+			else {
+				/* XXX: should never happen! */
+				rspamd_milter_send_task_results (nsession->client_milter_conn,
+						session->master_conn->results, NULL, 0);
+			}
+		}
+		else {
+			rspamd_milter_send_task_results (nsession->client_milter_conn,
+					session->master_conn->results, NULL, 0);
+		}
 		rspamd_http_message_free (msg);
 		REF_RELEASE (session);
 	}
@@ -1649,9 +1752,7 @@ rspamd_proxy_scan_self_reply (struct rspamd_task *task)
 				NULL,
 				ctype,
 				session,
-				session->client_sock,
-				NULL,
-				session->ctx->ev_base);
+				session->ctx->timeout / 10.0);
 	}
 }
 
@@ -1691,8 +1792,15 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 
 	msg = session->client_message;
 	task = rspamd_task_new (session->worker, session->ctx->cfg,
-			session->pool, session->ctx->lang_det);
+			session->pool, session->ctx->lang_det,
+			session->ctx->event_loop, FALSE);
 	task->flags |= RSPAMD_TASK_FLAG_MIME;
+
+	if (session->ctx->milter) {
+		task->protocol_flags |= RSPAMD_TASK_PROTOCOL_FLAG_MILTER|
+				RSPAMD_TASK_PROTOCOL_FLAG_BODY_BLOCK;
+	}
+
 	task->sock = -1;
 
 	if (session->client_milter_conn) {
@@ -1707,7 +1815,6 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 	task->resolver = session->ctx->resolver;
 	/* TODO: allow to disable autolearn in protocol */
 	task->flags |= RSPAMD_TASK_FLAG_LEARN_AUTO;
-	task->ev_base = session->ctx->ev_base;
 	task->s = rspamd_session_create (task->task_pool, rspamd_proxy_task_fin,
 			NULL, (event_finalizer_t )rspamd_task_free, task);
 	data = rspamd_http_message_get_body (msg, &len);
@@ -1737,13 +1844,21 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 
 	/* Set global timeout for the task */
 	if (session->ctx->default_upstream->timeout > 0.0) {
-		struct timeval task_tv;
+		task->timeout_ev.data = task;
+		ev_timer_init (&task->timeout_ev, rspamd_task_timeout,
+				session->ctx->default_upstream->timeout,
+				session->ctx->default_upstream->timeout);
+		ev_timer_start (task->event_loop, &task->timeout_ev);
 
-		event_set (&task->timeout_ev, -1, EV_TIMEOUT, rspamd_task_timeout,
-				task);
-		event_base_set (session->ctx->ev_base, &task->timeout_ev);
-		double_to_tv (session->ctx->default_upstream->timeout, &task_tv);
-		event_add (&task->timeout_ev, &task_tv);
+	}
+	else if (session->ctx->has_self_scan) {
+		if (session->ctx->cfg->task_timeout > 0) {
+			task->timeout_ev.data = task;
+			ev_timer_init (&task->timeout_ev, rspamd_task_timeout,
+					session->ctx->cfg->task_timeout,
+					session->ctx->default_upstream->timeout);
+			ev_timer_start (task->event_loop, &task->timeout_ev);
+		}
 	}
 
 	session->master_conn->task = task;
@@ -1796,9 +1911,26 @@ retry:
 			goto err;
 		}
 
-		session->master_conn->up = rspamd_upstream_get (backend->u,
-				RSPAMD_UPSTREAM_ROUND_ROBIN, NULL, 0);
-		session->master_conn->io_tv = &backend->io_tv;
+		/* Provide hash key if hashing based on source address is desired */
+		guint hash_len;
+		gpointer hash_key = rspamd_inet_address_get_hash_key (session->client_addr,
+				&hash_len);
+
+		if (session->ctx->max_retries > 1 &&
+			session->retries == session->ctx->max_retries) {
+
+			session->master_conn->up = rspamd_upstream_get_except (backend->u,
+					session->master_conn->up,
+					RSPAMD_UPSTREAM_ROUND_ROBIN,
+					hash_key, hash_len);
+		}
+		else {
+			session->master_conn->up = rspamd_upstream_get (backend->u,
+					RSPAMD_UPSTREAM_ROUND_ROBIN,
+					hash_key, hash_len);
+		}
+
+		session->master_conn->timeout = backend->timeout;
 
 		if (session->master_conn->up == NULL) {
 			msg_err_session ("cannot select upstream for %s",
@@ -1807,15 +1939,17 @@ retry:
 		}
 
 		session->master_conn->backend_sock = rspamd_inet_address_connect (
-				rspamd_upstream_addr (session->master_conn->up),
+				rspamd_upstream_addr_next (session->master_conn->up),
 				SOCK_STREAM, TRUE);
 
 		if (session->master_conn->backend_sock == -1) {
 			msg_err_session ("cannot connect upstream: %s(%s)",
 					host ? hostbuf : "default",
-							rspamd_inet_address_to_string (rspamd_upstream_addr (
-									session->master_conn->up)));
-			rspamd_upstream_fail (session->master_conn->up, TRUE);
+							rspamd_inet_address_to_string_pretty (
+									rspamd_upstream_addr_cur (
+											session->master_conn->up)));
+			rspamd_upstream_fail (session->master_conn->up, TRUE,
+					strerror (errno));
 			session->retries ++;
 			goto retry;
 		}
@@ -1832,22 +1966,19 @@ retry:
 			goto err; /* No fallback here */
 		}
 
-		session->master_conn->backend_conn = rspamd_http_connection_new (
+		session->master_conn->backend_conn = rspamd_http_connection_new_client_socket (
+				session->ctx->http_ctx,
 				NULL,
 				proxy_backend_master_error_handler,
 				proxy_backend_master_finish_handler,
 				RSPAMD_HTTP_CLIENT_SIMPLE,
-				RSPAMD_HTTP_CLIENT,
-				session->ctx->keys_cache,
-				NULL);
+				session->master_conn->backend_sock);
 		session->master_conn->flags &= ~RSPAMD_BACKEND_CLOSED;
 		session->master_conn->parser_from_ref = backend->parser_from_ref;
 		session->master_conn->parser_to_ref = backend->parser_to_ref;
 
 		if (backend->key) {
 			msg->peer_key = rspamd_pubkey_ref (backend->key);
-			rspamd_http_connection_set_key (session->master_conn->backend_conn,
-					session->ctx->local_key);
 		}
 
 		if (backend->settings_id != NULL) {
@@ -1858,7 +1989,8 @@ retry:
 
 		if (backend->local ||
 				rspamd_inet_address_is_local (
-						rspamd_upstream_addr (session->master_conn->up), FALSE)) {
+						rspamd_upstream_addr_cur (
+								session->master_conn->up))) {
 
 			if (session->fname) {
 				rspamd_http_message_add_header (msg, "File", session->fname);
@@ -1869,8 +2001,7 @@ retry:
 			rspamd_http_connection_write_message_shared (
 					session->master_conn->backend_conn,
 					msg, NULL, NULL, session->master_conn,
-					session->master_conn->backend_sock,
-					session->master_conn->io_tv, session->ctx->ev_base);
+					session->master_conn->timeout);
 		}
 		else {
 			if (session->fname) {
@@ -1898,8 +2029,7 @@ retry:
 			rspamd_http_connection_write_message (
 					session->master_conn->backend_conn,
 					msg, NULL, NULL, session->master_conn,
-					session->master_conn->backend_sock,
-					session->master_conn->io_tv, session->ctx->ev_base);
+					session->master_conn->timeout);
 		}
 	}
 
@@ -2035,37 +2165,51 @@ proxy_milter_finish_handler (gint fd,
 
 static void
 proxy_milter_error_handler (gint fd,
-		struct rspamd_milter_session *rms,
+		struct rspamd_milter_session *rms, /* unused */
 		void *ud, GError *err)
 {
 	struct rspamd_proxy_session *session = ud;
 
-	msg_info_session ("abnormally closing milter connection from: %s, "
-			"error: %s", rspamd_inet_address_to_string (session->client_addr),
-			err->message);
-	/* Terminate session immediately */
-	proxy_backend_close_connection (session->master_conn);
-	REF_RELEASE (session);
+	if (err && err->code != 0) {
+		msg_info_session ("abnormally closing milter connection from: %s, "
+						  "error: %e",
+				rspamd_inet_address_to_string_pretty (session->client_addr),
+				err);
+		/* Terminate session immediately */
+		proxy_backend_close_connection (session->master_conn);
+		REF_RELEASE (session);
+	}
+	else {
+		msg_info_session ("normally closing milter connection from: %s, "
+						  "%e",
+				rspamd_inet_address_to_string_pretty (session->client_addr),
+				err);
+		/* Terminate session immediately */
+		proxy_backend_close_connection (session->master_conn);
+		REF_RELEASE (session);
+	}
 }
 
 static void
-proxy_accept_socket (gint fd, short what, void *arg)
+proxy_accept_socket (EV_P_ ev_io *w, int revents)
 {
-	struct rspamd_worker *worker = (struct rspamd_worker *) arg;
+	struct rspamd_worker *worker = (struct rspamd_worker *)w->data;
 	struct rspamd_proxy_ctx *ctx;
-	rspamd_inet_addr_t *addr;
+	rspamd_inet_addr_t *addr = NULL;
 	struct rspamd_proxy_session *session;
 	gint nfd;
 
 	ctx = worker->ctx;
 
 	if ((nfd =
-		rspamd_accept_from_socket (fd, &addr, worker->accept_events)) == -1) {
+		rspamd_accept_from_socket (w->fd, &addr,
+				rspamd_worker_throttle_accept_events, worker->accept_events)) == -1) {
 		msg_warn ("accept failed: %s", strerror (errno));
 		return;
 	}
 	/* Check for EAGAIN */
 	if (nfd == 0) {
+		rspamd_inet_address_free (addr);
 		return;
 	}
 
@@ -2076,7 +2220,7 @@ proxy_accept_socket (gint fd, short what, void *arg)
 	session->mirror_conns = g_ptr_array_sized_new (ctx->mirrors->len);
 
 	session->pool = rspamd_mempool_new (rspamd_mempool_suggest_size (),
-			"proxy");
+			"proxy", 0);
 	session->ctx = ctx;
 	session->worker = worker;
 
@@ -2086,13 +2230,13 @@ proxy_accept_socket (gint fd, short what, void *arg)
 	}
 
 	if (!ctx->milter) {
-		session->client_conn = rspamd_http_connection_new (NULL,
+		session->client_conn = rspamd_http_connection_new_server (
+				ctx->http_ctx,
+				nfd,
+				NULL,
 				proxy_client_error_handler,
 				proxy_client_finish_handler,
-				0,
-				RSPAMD_HTTP_SERVER,
-				ctx->keys_cache,
-				NULL);
+				0);
 
 		if (ctx->key) {
 			rspamd_http_connection_set_key (session->client_conn, ctx->key);
@@ -2104,9 +2248,7 @@ proxy_accept_socket (gint fd, short what, void *arg)
 
 		rspamd_http_connection_read_message_shared (session->client_conn,
 				session,
-				nfd,
-				&ctx->io_tv,
-				ctx->ev_base);
+				session->ctx->timeout);
 	}
 	else {
 		msg_info_session ("accepted milter connection from %s port %d",
@@ -2130,9 +2272,9 @@ proxy_accept_socket (gint fd, short what, void *arg)
 		}
 #endif
 
-		rspamd_milter_handle_socket (nfd, NULL,
+		rspamd_milter_handle_socket (nfd, 0.0,
 				session->pool,
-				ctx->ev_base,
+				ctx->event_loop,
 				proxy_milter_finish_handler,
 				proxy_milter_error_handler,
 				session);
@@ -2140,61 +2282,104 @@ proxy_accept_socket (gint fd, short what, void *arg)
 }
 
 static void
-proxy_rotate_key (gint fd, short what, void *arg)
+adjust_upstreams_limits (struct rspamd_proxy_ctx *ctx)
 {
-	struct timeval rot_tv;
-	struct rspamd_proxy_ctx *ctx = arg;
-	gpointer kp;
+	struct rspamd_http_upstream *backend;
+	gpointer k, v;
+	GHashTableIter it;
 
-	double_to_tv (ctx->rotate_tm, &rot_tv);
-	rot_tv.tv_sec += ottery_rand_range (rot_tv.tv_sec);
-	event_del (&ctx->rotate_ev);
-	event_add (&ctx->rotate_ev, &rot_tv);
+	/*
+	 * We set error time equal to max_retries * backend_timeout and max_errors
+	 * to max_retries - 1
+	 *
+	 * So if we failed to scan a message on a backend for some reasons, we
+	 * will try to re-resolve it faster
+	 */
 
-	kp = ctx->local_key;
-	ctx->local_key = rspamd_keypair_new (RSPAMD_KEYPAIR_KEX,
-			RSPAMD_CRYPTOBOX_MODE_25519);
-	rspamd_keypair_unref (kp);
+	g_hash_table_iter_init (&it, ctx->upstreams);
+
+	while (g_hash_table_iter_next (&it, &k, &v)) {
+		backend = (struct rspamd_http_upstream *)v;
+
+		if (!backend->self_scan && backend->u) {
+			rspamd_upstreams_set_limits (backend->u,
+					NAN, NAN, ctx->max_retries * backend->timeout, NAN,
+					ctx->max_retries - 1, 0);
+		}
+	}
 }
 
+__attribute__((noreturn))
 void
-start_rspamd_proxy (struct rspamd_worker *worker) {
+start_rspamd_proxy (struct rspamd_worker *worker)
+{
 	struct rspamd_proxy_ctx *ctx = worker->ctx;
-	struct timeval rot_tv;
+	gboolean is_controller = FALSE;
 
+	g_assert (rspamd_worker_check_context (worker->ctx, rspamd_rspamd_proxy_magic));
 	ctx->cfg = worker->srv->cfg;
-	ctx->ev_base = rspamd_prepare_worker (worker, "rspamd_proxy",
+	ctx->event_loop = rspamd_prepare_worker (worker, "rspamd_proxy",
 			proxy_accept_socket);
 
-	ctx->resolver = dns_resolver_init (worker->srv->logger,
-			ctx->ev_base,
+	ctx->resolver = rspamd_dns_resolver_init (worker->srv->logger,
+			ctx->event_loop,
 			worker->srv->cfg);
-	double_to_tv (ctx->timeout, &ctx->io_tv);
-	rspamd_map_watch (worker->srv->cfg, ctx->ev_base, ctx->resolver, worker, 0);
 
 	rspamd_upstreams_library_config (worker->srv->cfg, ctx->cfg->ups_ctx,
-			ctx->ev_base, ctx->resolver->r);
+			ctx->event_loop, ctx->resolver->r);
 
-	/* XXX: stupid default */
-	ctx->keys_cache = rspamd_keypair_cache_new (256);
-	ctx->local_key = rspamd_keypair_new (RSPAMD_KEYPAIR_KEX,
-			RSPAMD_CRYPTOBOX_MODE_25519);
-
-	double_to_tv (ctx->rotate_tm, &rot_tv);
-	rot_tv.tv_sec += ottery_rand_range (rot_tv.tv_sec);
-	event_set (&ctx->rotate_ev, -1, EV_TIMEOUT, proxy_rotate_key, ctx);
-	event_base_set (ctx->ev_base, &ctx->rotate_ev);
-	event_add (&ctx->rotate_ev, &rot_tv);
+	ctx->http_ctx = rspamd_http_context_create (ctx->cfg, ctx->event_loop,
+			ctx->cfg->ups_ctx);
+	rspamd_mempool_add_destructor (ctx->cfg->cfg_pool,
+			(rspamd_mempool_destruct_t)rspamd_http_context_free,
+			ctx->http_ctx);
 
 	if (ctx->has_self_scan) {
 		/* Additional initialisation needed */
-		rspamd_worker_init_scanner (worker, ctx->ev_base, ctx->resolver,
+		rspamd_worker_init_scanner (worker, ctx->event_loop, ctx->resolver,
 				&ctx->lang_det);
+
+		if (worker->index == 0) {
+			/*
+			 * If there are no controllers and no normal workers,
+			 * then pretend that we are a controller
+			 */
+			gboolean controller_seen = FALSE;
+			GList *cur;
+
+			cur = worker->srv->cfg->workers;
+
+			while (cur) {
+				struct rspamd_worker_conf *cf;
+
+				cf = (struct rspamd_worker_conf *)cur->data;
+				if ((cf->type == g_quark_from_static_string ("controller")) ||
+						(cf->type == g_quark_from_static_string ("normal"))) {
+
+					if (cf->enabled && cf->count >= 0) {
+						controller_seen = TRUE;
+						break;
+					}
+				}
+
+				cur = g_list_next (cur);
+			}
+
+			if (!controller_seen) {
+				msg_info ("no controller or normal workers defined, execute "
+							  "controller periodics in this worker");
+				worker->flags |= RSPAMD_WORKER_CONTROLLER;
+				is_controller = TRUE;
+			}
+		}
+	}
+	else {
+		worker->flags &= ~RSPAMD_WORKER_SCANNER;
 	}
 
 	if (worker->srv->cfg->enable_sessions_cache) {
 		ctx->sessions_cache = rspamd_worker_session_cache_new (worker,
-				ctx->ev_base);
+				ctx->event_loop);
 	}
 
 	ctx->milter_ctx.spam_header = ctx->spam_header;
@@ -2203,21 +2388,40 @@ start_rspamd_proxy (struct rspamd_worker *worker) {
 	ctx->milter_ctx.sessions_cache = ctx->sessions_cache;
 	ctx->milter_ctx.client_ca_name = ctx->client_ca_name;
 	ctx->milter_ctx.reject_message = ctx->reject_message;
+	ctx->milter_ctx.cfg = ctx->cfg;
 	rspamd_milter_init_library (&ctx->milter_ctx);
 
-	rspamd_lua_run_postloads (ctx->cfg->lua_state, ctx->cfg, ctx->ev_base,
-			worker);
+	if (is_controller) {
+		rspamd_worker_init_controller (worker, NULL);
+	}
+	else {
+		if (ctx->has_self_scan) {
+			rspamd_map_watch (worker->srv->cfg, ctx->event_loop, ctx->resolver,
+					worker, RSPAMD_MAP_WATCH_SCANNER);
+		}
+		else {
+			rspamd_map_watch (worker->srv->cfg, ctx->event_loop, ctx->resolver,
+					worker, RSPAMD_MAP_WATCH_WORKER);
+		}
+	}
 
-	event_base_loop (ctx->ev_base, 0);
+	rspamd_lua_run_postloads (ctx->cfg->lua_state, ctx->cfg, ctx->event_loop,
+			worker);
+	adjust_upstreams_limits (ctx);
+
+	ev_loop (ctx->event_loop, 0);
 	rspamd_worker_block_signals ();
 
 	if (ctx->has_self_scan) {
 		rspamd_stat_close ();
 	}
 
-	rspamd_keypair_cache_destroy (ctx->keys_cache);
+	if (is_controller) {
+		rspamd_controller_on_terminate (worker, NULL);
+	}
+
 	REF_RELEASE (ctx->cfg);
-	rspamd_log_close (worker->srv->logger, TRUE);
+	rspamd_log_close (worker->srv->logger);
 
 	exit (EXIT_SUCCESS);
 }
